# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from typing import Any

import pandas as pd

from allotropy.allotrope.models.pcr_benchling_2023_09_dpcr import (
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    DPCRAggregateDocument,
    DPCRDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    TCalculatedDataAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
    TQuantityValueNumberPerMicroliter,
)
from allotropy.allotrope.models.shared.definitions.definitions import TQuantityValue
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_reader import AbsoluteQReader
from allotropy.parsers.appbio_absolute_q.constants import (
    AGGREGATION_LOOKUP,
    CALCULATED_DATA_REFERENCE,
    CalculatedDataItem,
    CalculatedDataSource,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class AppbioAbsoluteQParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        raw_contents = named_file_contents.contents
        filename = named_file_contents.original_file_name
        reader = AbsoluteQReader(raw_contents)
        return self._get_model(reader.wells, reader.group_rows, filename)

    def _get_model(
        self, wells: pd.DataFrame, group_rows: pd.DataFrame, filename: str
    ) -> Model:
        well_groups = wells.groupby(["Well"]).groups.keys()
        group_ids: dict[Any, list] = defaultdict(list)

        dpcr_document = [
            self._get_dpcr_document(wells[wells["Well"] == well_name], group_ids)
            for well_name in well_groups
        ]

        calculated_data_aggregate_document = None
        calculated_data_document = self.get_calculated_data_document(
            group_ids, group_rows
        )

        if calculated_data_document:
            calculated_data_aggregate_document = TCalculatedDataAggregateDocument(
                calculated_data_document=calculated_data_document
            )

        return Model(
            dPCR_aggregate_document=DPCRAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=wells.iloc[0]["Instrument"],
                    brand_name="QuantStudio Absolute Q Digital PCR System",
                    product_manufacturer="ThermoFisher Scientific",
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name="QuantStudio Absolute Q Digital PCR Software",
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                dPCR_document=dpcr_document,
                calculated_data_aggregate_document=calculated_data_aggregate_document,
            )
        )

    def _get_dpcr_document(
        self, well_data: pd.DataFrame, group_ids: dict[Any, list]
    ) -> DPCRDocumentItem:
        measurement_documents = []
        for _, well_item in well_data.iterrows():
            measurement_identifier = random_uuid_str()

            key = str((well_item["Group"], well_item["Target"]))
            group_ids[key].append(measurement_identifier)

            measurement_documents.append(
                MeasurementDocumentItem(
                    measurement_identifier=measurement_identifier,
                    measurement_time=self._get_date_time(str(well_item["Date"])),
                    target_DNA_description=well_item["Target"],
                    total_partition_count=TQuantityValueNumber(
                        value=well_item["Total"]
                    ),
                    sample_document=SampleDocument(
                        sample_identifier=well_item["Name"],
                        well_location_identifier=well_item["Well"],
                        well_plate_identifier=well_item["Plate"],
                    ),
                    device_control_aggregate_document=DeviceControlAggregateDocument(
                        device_control_document=[
                            DeviceControlDocumentItem(
                                device_type="dPCR",
                                reporter_dye_setting=well_item["Dye"],
                            )
                        ]
                    ),
                    processed_data_aggregate_document=ProcessedDataAggregateDocument(
                        processed_data_document=[
                            ProcessedDataDocumentItem(
                                number_concentration=TQuantityValueNumberPerMicroliter(
                                    value=well_item["Conc. cp/uL"]
                                ),
                                positive_partition_count=TQuantityValueNumber(
                                    value=well_item["Positives"]
                                ),
                            )
                        ]
                    ),
                )
            )
        return DPCRDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                experimental_data_identifier=well_data.iloc[0]["Run"],
                plate_well_count=TQuantityValueNumber(value=16),
                container_type=ContainerType.well_plate,
                measurement_document=measurement_documents,
            )
        )

    @staticmethod
    def get_calculated_data_document(
        group_ids: dict, group_rows: pd.DataFrame
    ) -> list[CalculatedDataDocumentItem]:
        calculated_data_document: list[CalculatedDataDocumentItem] = []

        for _, group in group_rows.iterrows():
            aggregation_type = AGGREGATION_LOOKUP[group["Well"]]

            # Samples designated as "Individual" include no calculated data
            if aggregation_type not in CALCULATED_DATA_REFERENCE:
                continue

            group_name = group["Group"].split("(")[0].strip()
            key = str((group_name, group["Target"]))

            ids = group_ids[key]
            calculated_data_ids = {}
            defered_calculated_data_items: list[CalculatedDataItem] = []

            for calculated_data_item in CALCULATED_DATA_REFERENCE[aggregation_type]:
                # TODO: if aggregation type is Replicate(Average), check for required columns
                # Raise if column(s) do not exist

                # Calculated data items that have another calculated data as data source
                # need to be created after the later
                if calculated_data_item.source == CalculatedDataSource.CALCULATED_DATA:
                    defered_calculated_data_items.append(calculated_data_item)
                    continue

                datum_value = float(group[calculated_data_item.column])
                calculated_data_id = random_uuid_str()
                calculated_data_ids[calculated_data_item.name] = calculated_data_id

                data_source_document = [
                    DataSourceDocumentItem(
                        data_source_identifier=identifier,
                        data_source_feature=calculated_data_item.source_feature,
                    )
                    for identifier in ids
                ]
                calculated_data_document.append(
                    CalculatedDataDocumentItem(
                        calculated_data_identifier=calculated_data_id,
                        data_source_aggregate_document=DataSourceAggregateDocument(
                            data_source_document=data_source_document
                        ),
                        calculated_data_name=calculated_data_item.name,
                        calculated_datum=TQuantityValue(
                            value=datum_value,
                            unit=calculated_data_item.unit,
                        ),
                    )
                )

            # TODO: this should be inproved (repeat less code)
            for calculated_data_item in defered_calculated_data_items:
                datum_value = float(group[calculated_data_item.column])
                calculated_data_id = random_uuid_str()

                data_source_document = [
                    DataSourceDocumentItem(
                        data_source_identifier=calculated_data_ids[source_feature],
                        data_source_feature=source_feature,
                    )
                    for source_feature in calculated_data_item.source_feature.split(",")
                ]

                calculated_data_document.append(
                    CalculatedDataDocumentItem(
                        calculated_data_identifier=calculated_data_id,
                        data_source_aggregate_document=DataSourceAggregateDocument(
                            data_source_document=data_source_document
                        ),
                        calculated_data_name=calculated_data_item.name,
                        calculated_datum=TQuantityValue(
                            value=datum_value,
                            unit=calculated_data_item.unit,
                        ),
                    )
                )

        return calculated_data_document



from datetime import tzinfo
from enum import Enum
from typing import Optional

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.agilent_gen5.agilent_gen5_parser import AgilentGen5Parser
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_parser import (
    AppbioAbsoluteQParser,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_parser import (
    AppBioQuantStudioParser,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_parser import (
    AppBioQuantStudioDesignandanalysisParser,
)
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_parser import ViCellBluParser
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_parser import ViCellXRParser
from allotropy.parsers.chemometec_nucleoview.nucleoview_parser import (
    ChemometecNucleoviewParser,
)
from allotropy.parsers.example_weyland_yutani.example_weyland_yutani_parser import (
    ExampleWeylandYutaniParser,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_parser import (
    LuminexXponentParser,
)
from allotropy.parsers.moldev_softmax_pro.softmax_pro_parser import SoftmaxproParser
from allotropy.parsers.novabio_flex2.novabio_flex2_parser import NovaBioFlexParser
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_parser import (
    PerkinElmerEnvisionParser,
)
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_parser import QiacuitydPCRParser
from allotropy.parsers.revvity_kaleido.kaleido_parser import KaleidoParser
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_parser import (
    RocheCedexBiohtParser,
)
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_parser import (
    NanodropEightParser,
)
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_parser import (
    UnchainedLabsLunaticParser,
)
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.vendor_parser import VendorParser


class Vendor(Enum):
    AGILENT_GEN5 = "AGILENT_GEN5"
    APPBIO_ABSOLUTE_Q = "APPBIO_ABSOLUTE_Q"
    APPBIO_QUANTSTUDIO = "APPBIO_QUANTSTUDIO"
    APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS = "APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS"
    BECKMAN_VI_CELL_BLU = "BECKMAN_VI_CELL_BLU"
    BECKMAN_VI_CELL_XR = "BECKMAN_VI_CELL_XR"
    CHEMOMETEC_NUCLEOVIEW = "CHEMOMETEC_NUCLEOVIEW"
    EXAMPLE_WEYLAND_YUTANI = "EXAMPLE_WEYLAND_YUTANI"
    LUMINEX_XPONENT = "LUMINEX_XPONENT"
    MOLDEV_SOFTMAX_PRO = "MOLDEV_SOFTMAX_PRO"
    NOVABIO_FLEX2 = "NOVABIO_FLEX2"
    PERKIN_ELMER_ENVISION = "PERKIN_ELMER_ENVISION"
    QIACUITY_DPCR = "QIACUITY_DPCR"
    REVVITY_KALEIDO = "REVVITY_KALEIDO"
    ROCHE_CEDEX_BIOHT = "ROCHE_CEDEX_BIOHT"
    THERMO_FISHER_NANODROP_EIGHT = "THERMO_FISHER_NANODROP_EIGHT"
    UNCHAINED_LABS_LUNATIC = "UNCHAINED_LABS_LUNATIC"


_VENDOR_TO_PARSER: dict[Vendor, type[VendorParser]] = {
    Vendor.AGILENT_GEN5: AgilentGen5Parser,
    Vendor.APPBIO_ABSOLUTE_Q: AppbioAbsoluteQParser,
    Vendor.APPBIO_QUANTSTUDIO: AppBioQuantStudioParser,
    Vendor.APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS: AppBioQuantStudioDesignandanalysisParser,
    Vendor.BECKMAN_VI_CELL_BLU: ViCellBluParser,
    Vendor.BECKMAN_VI_CELL_XR: ViCellXRParser,
    Vendor.CHEMOMETEC_NUCLEOVIEW: ChemometecNucleoviewParser,
    Vendor.EXAMPLE_WEYLAND_YUTANI: ExampleWeylandYutaniParser,
    Vendor.LUMINEX_XPONENT: LuminexXponentParser,
    Vendor.MOLDEV_SOFTMAX_PRO: SoftmaxproParser,
    Vendor.NOVABIO_FLEX2: NovaBioFlexParser,
    Vendor.PERKIN_ELMER_ENVISION: PerkinElmerEnvisionParser,
    Vendor.QIACUITY_DPCR: QiacuitydPCRParser,
    Vendor.REVVITY_KALEIDO: KaleidoParser,
    Vendor.ROCHE_CEDEX_BIOHT: RocheCedexBiohtParser,
    Vendor.THERMO_FISHER_NANODROP_EIGHT: NanodropEightParser,
    Vendor.UNCHAINED_LABS_LUNATIC: UnchainedLabsLunaticParser,
}


def get_parser(
    vendor: Vendor, default_timezone: Optional[tzinfo] = None
) -> VendorParser:
    timestamp_parser = TimestampParser(default_timezone)
    try:
        return _VENDOR_TO_PARSER[vendor](timestamp_parser)
    except KeyError as e:
        error = f"Failed to create parser, unregistered vendor: {vendor}."
        raise AllotropeConversionError(error) from e



# generated by datamodel-codegen:
#   filename:  multi-analyte-profiling.json
#   timestamp: 2024-01-26T15:48:36+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional, Union

from allotropy.allotrope.models.shared.components.plate_reader import (
    SampleRoleType,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDateTimeStampValue,
    TQuantityValue,
    TStringValue,
)


@dataclass
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue
    field_index: Optional[int] = None


@dataclass
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass
class Manifest:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: Optional[str] = None
    field_type: Optional[str] = None
    shapes: Optional[list[str]] = None


@dataclass
class DeviceDocumentItem:
    device_identifier: Optional[TStringValue] = None
    device_type: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    product_manufacturer: Optional[TStringValue] = None
    brand_name: Optional[TStringValue] = None
    equipment_serial_number: Optional[TStringValue] = None
    firmware_version: Optional[TStringValue] = None
    field_index: Optional[int] = None


@dataclass
class ReferenceMaterialDocument:
    reference_material_identifier: Optional[TStringValue] = None
    batch_identifier: Optional[TStringValue] = None
    expiry_time_prescription: Optional[TDateTimeStampValue] = None


@dataclass
class CalibrationResultDocumentItem:
    calibration_result_name: Optional[TStringValue] = None
    calibration_result: Optional[TQuantityValueUnitless] = None


@dataclass
class CalibrationResultAggregateDocument:
    calibration_result_document: Optional[list[CalibrationResultDocumentItem]] = None


@dataclass
class CalibrationDocumentItem:
    calibration_name: Optional[TStringValue] = None
    calibration_description: Optional[TStringValue] = None
    calibration_time: Optional[TDateTimeStampValue] = None
    expiry_time_prescription: Optional[TDateTimeStampValue] = None
    calibration_report: Optional[TStringValue] = None
    reference_material_document: Optional[ReferenceMaterialDocument] = None
    calibration_result_aggregate_document: Optional[
        CalibrationResultAggregateDocument
    ] = None


@dataclass
class CalibrationAggregateDocument:
    calibration_document: Optional[list[CalibrationDocumentItem]] = None


@dataclass
class DeviceSystemDocument:
    asset_management_identifier: Optional[TStringValue] = None
    description: Optional[Any] = None
    brand_name: Optional[TStringValue] = None
    product_manufacturer: Optional[TStringValue] = None
    device_identifier: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    equipment_serial_number: Optional[TStringValue] = None
    firmware_version: Optional[TStringValue] = None
    device_document: Optional[list[DeviceDocumentItem]] = None
    calibration_aggregate_document: Optional[CalibrationAggregateDocument] = None


@dataclass
class DataSystemDocument:
    data_system_instance_identifier: Optional[TStringValue] = None
    file_name: Optional[TStringValue] = None
    UNC_path: Optional[TStringValue] = None
    software_name: Optional[TStringValue] = None
    software_version: Optional[TStringValue] = None
    ASM_converter_name: Optional[TStringValue] = None
    ASM_converter_version: Optional[TStringValue] = None


@dataclass
class SampleDocument:
    sample_identifier: TStringValue
    description: Optional[Any] = None
    batch_identifier: Optional[TStringValue] = None
    group_identifier: Optional[TStringValue] = None
    sample_role_type: Optional[SampleRoleType] = None
    written_name: Optional[TStringValue] = None
    location_identifier: Optional[TStringValue] = None
    well_plate_identifier: Optional[TStringValue] = None


@dataclass
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    product_manufacturer: Optional[TStringValue] = None
    brand_name: Optional[TStringValue] = None
    equipment_serial_number: Optional[TStringValue] = None
    firmware_version: Optional[TStringValue] = None
    sample_volume_setting: Optional[TQuantityValueMicroliter] = None
    dilution_factor_setting: Optional[TQuantityValueUnitless] = None
    detector_gain_setting: Optional[TStringValue] = None
    minimum_assay_bead_count_setting: Optional[TQuantityValueUnitless] = None


@dataclass
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass
class AnalyteDocumentItem:
    analyte_identifier: TStringValue
    analyte_name: TStringValue
    assay_bead_identifier: TStringValue
    assay_bead_count: TQuantityValueNumber
    fluorescence: TQuantityValueRelativeFluorescenceUnit


@dataclass
class AnalyteAggregateDocument:
    analyte_document: list[AnalyteDocumentItem]


@dataclass
class ErrorDocumentItem:
    error: TStringValue
    error_feature: Optional[TStringValue] = None


@dataclass
class ErrorAggregateDocument:
    error_document: Optional[list[ErrorDocumentItem]] = None


@dataclass
class MeasurementDocumentItem:
    measurement_identifier: TStringValue
    measurement_time: TDateTimeStampValue
    sample_document: SampleDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    assay_bead_count: TQuantityValueNumber
    analyte_aggregate_document: AnalyteAggregateDocument
    error_aggregate_document: Optional[ErrorAggregateDocument] = None


@dataclass
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]
    analytical_method_identifier: Optional[TStringValue] = None
    method_version: Optional[TStringValue] = None
    experimental_data_identifier: Optional[TStringValue] = None
    experiment_type: Optional[TStringValue] = None
    container_type: Optional[TStringValue] = None
    plate_well_count: Optional[TQuantityValueNumber] = None


@dataclass
class MultiAnalyteProfilingDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: Optional[TStringValue] = None
    submitter: Optional[TStringValue] = None


@dataclass
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValue
    data_source_aggregate_document: Optional[DataSourceAggregateDocument] = None
    calculated_data_identifier: Optional[TStringValue] = None
    calculation_description: Optional[TStringValue] = None
    field_index: Optional[int] = None


@dataclass
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass
class Asm:
    field_asm_manifest: Optional[Union[str, Manifest]] = None


@dataclass
class MultiAnalyteProfilingAggregateDocument:
    device_system_document: DeviceSystemDocument
    multi_analyte_profiling_document: list[MultiAnalyteProfilingDocumentItem]
    data_system_document: Optional[DataSystemDocument] = None
    calculated_data_aggregate_document: Optional[CalculatedDataAggregateDocument] = None


@dataclass
class Model(Asm):
    field_asm_manifest: Union[str, Manifest]
    multi_analyte_profiling_aggregate_document: Optional[
        MultiAnalyteProfilingAggregateDocument
    ] = None



from __future__ import annotations

from typing import Optional

import numpy as np

from allotropy.allotrope.models.plate_reader_benchling_2023_09_plate_reader import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueNanometer,
    TQuantityValueNumber,
)
from allotropy.allotrope.models.shared.definitions.definitions import TQuantityValue
from allotropy.allotrope.pandas_util import read_csv
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_structure import (
    Data,
    Measurement,
    WellPlate,
)
from allotropy.parsers.vendor_parser import VendorParser


class UnchainedLabsLunaticParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        raw_contents = named_file_contents.contents
        data = read_csv(filepath_or_buffer=raw_contents).replace(np.nan, None)

        filename = named_file_contents.original_file_name
        return self._get_model(Data.create(data), filename)

    def _get_model(self, data: Data, filename: str) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.device_identifier,
                    model_number="Lunatic",
                    product_manufacturer="Unchained Labs",
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name="Lunatic and Stunner Analysis",
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    self._get_plate_reader_document_item(plate)
                    for plate in data.well_plate_list
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data
                ),
            ),
        )

    def _get_plate_reader_document_item(
        self, plate: WellPlate
    ) -> PlateReaderDocumentItem:
        return PlateReaderDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=self._get_date_time(plate.measurement_time),
                analytical_method_identifier=plate.analytical_method_identifier,
                container_type=ContainerType.well_plate,
                plate_well_count=TQuantityValueNumber(value=96),
                measurement_document=[
                    self._get_measurement_document_item(measurement)
                    for measurement in plate.measurements
                ],
            )
        )

    def _get_measurement_document_item(
        self, measurement: Measurement
    ) -> UltravioletAbsorbancePointDetectionMeasurementDocumentItems:
        return UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                        device_type="plate reader",
                        detector_wavelength_setting=TQuantityValueNanometer(
                            value=measurement.wavelength
                        ),
                    )
                ]
            ),
            absorbance=TQuantityValueMilliAbsorbanceUnit(value=measurement.absorbance),
            sample_document=SampleDocument(
                sample_identifier=measurement.sample_identifier,
                location_identifier=measurement.location_identifier,
                well_plate_identifier=measurement.well_plate_identifier,
            ),
        )

    def _get_calculated_data_aggregate_document(
        self, data: Data
    ) -> Optional[CalculatedDataAggregateDocument]:
        if not (calculated_data_document := data.get_calculated_data_document()):
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculated_result=TQuantityValue(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_source_document
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_document
            ]
        )



#!/usr/bin/env python

# Disable warnings about use of random.random with SHA256.
# ruff: noqa: S311

"""Generate random plates with some formatting gotchas.

Default output is CSV that looks like:

    Weyland-Yutani 470 1879,,,,
    Recorded,2023-10-23:08:21:45,,,
    ,,,,
    ,A,B,C,D
    1,1.67,2.35,4.17,2.75
    2,2.78,2.52,3.23,1.81
    3,0.91,1.35,3.73,1.50
    4,3.34,3.46,3.70,0.77

In order:

-   Machine model and serial number
-   Recording timestamp
-   Blank line
-   4x4 table of readings
-   Blank line

Options:

-   `--checksum`: add footer row with SHA256 checksum
-   `--empty`: create some empty reading cells
-   `--exptime`: specify experiment timestamp
-   `--height`: specify number of rows
-   `--operator`: add a header row with operator name
-   `--out`: specify output file
-   `--seed`: specify random number seed
-   `--serial`: specify machine serial number
-   `--width`: specify number of columns

"""

import argparse
import csv
from datetime import datetime
import hashlib
import random
import sys
from typing import Any

import pytz

CHECKSUM_RANGE = 2**16
MODEL = "Weyland-Yutani 470"
PLATE_HEIGHT = 4
PLATE_WIDTH = 4
SERIAL_NUMBER_RANGE = 1000
MAX_READING = 5.0


def main() -> None:
    """Main driver."""
    args = parse_args()
    head = head_generate(args)
    body = body_generate(args)
    foot = foot_generate(args, body)
    result = csv_normalize([*head, *body, *foot])
    save(args, result)


def body_add_titles(args: argparse.Namespace, readings: list[Any]) -> list[Any]:
    """Make column titles for plate readings."""
    title_row = ["", *[chr(ord("A") + col) for col in range(args.width)]]
    readings = [[str(i + 1), *r] for (i, r) in enumerate(readings)]
    return [title_row, *readings]


def body_calculate_checksum(body: list[Any]) -> str:
    """Calculate SHA256 checksum for body."""
    m = hashlib.sha256()
    for row in body:
        for val in row:
            m.update(bytes(val, "utf-8"))
    digest = int(m.hexdigest(), base=16) % CHECKSUM_RANGE
    return f"{digest:04x}"


def body_empty_readings(args: argparse.Namespace, readings: list[Any]) -> list[Any]:
    """Replace some readings with empty."""
    product = args.width * args.height
    coords = list(range(product))
    coords = random.sample(coords, args.empty)
    for c in coords:
        col = c % args.width
        row = c // args.width
        readings[row][col] = ""
    return readings


def body_generate(args: argparse.Namespace) -> list[Any]:
    """Make body of plate."""
    readings = body_generate_readings(args)
    readings = body_empty_readings(args, readings)
    readings = body_add_titles(args, readings)
    return readings


def body_generate_readings(args: argparse.Namespace) -> list[Any]:
    """Make table of plate readings."""
    return [
        [f"{(random.random() * MAX_READING):.02f}" for _ in range(args.width)]
        for row in range(args.height)
    ]


def csv_normalize(rows: list[Any]) -> list[Any]:
    required = max(len(r) for r in rows)
    for row in rows:
        row.extend([""] * (required - len(row)))
    return rows


def foot_generate(args: argparse.Namespace, body: list[Any]) -> list[Any]:
    """Make foot of plate."""
    result = []
    if args.checksum:
        result.extend(
            [
                [],
                ["Checksum", body_calculate_checksum(body)],
            ]
        )
    return result


def head_generate(args: argparse.Namespace) -> list[Any]:
    """Make head of plate."""
    operator = [["Operator", args.operator]] if args.operator else []
    return [
        [f"{MODEL} {args.serial}"],
        ["Recorded", args.exptime],
        *operator,
        [],
    ]


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--checksum", action="store_true", default=False, help="include checksum"
    )
    parser.add_argument("--empty", type=int, default=0, help="number of empty readings")
    parser.add_argument("--exptime", type=str, default=None, help="reading date/time")
    parser.add_argument("--height", type=int, default=PLATE_HEIGHT, help="plate height")
    parser.add_argument("--operator", type=str, default=None, help="operator name")
    parser.add_argument("--out", type=str, default=None, help="output file")
    parser.add_argument("--seed", type=int, default=None, help="RNG seed")
    parser.add_argument("--serial", type=str, default=None, help="Reader serial number")
    parser.add_argument("--width", type=int, default=PLATE_WIDTH, help="plate width")
    args = parser.parse_args()

    if args.seed is not None:
        random.seed(args.seed)

    if args.serial is None:
        args.serial = SERIAL_NUMBER_RANGE + random.randint(1, SERIAL_NUMBER_RANGE - 1)

    if args.exptime is None:
        args.exptime = datetime.now(tz=pytz.utc).strftime("%Y-%m-%d:%H:%M:%S")

    return args


def save(args: argparse.Namespace, rows: list[Any]) -> None:
    """Save as CSV."""
    if args.out is None:
        csv.writer(sys.stdout).writerows(rows)
    else:
        with open(args.out, "w") as writer:
            csv.writer(writer).writerows(rows)


if __name__ == "__main__":
    main()



from typing import Optional

from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import (
    BaselineCorrectedReporterDataCube,
    ContainerType,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    MeltingCurveDataCube,
    Model,
    NormalizedReporterDataCube,
    PassiveReferenceDyeDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    QPCRAggregateDocument,
    QPCRDocumentItem,
    ReporterDyeDataCube,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueUnitless,
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_contents import (
    DesignQuantstudioContents,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_structure import (
    Data,
    Well,
    WellItem,
)
from allotropy.parsers.vendor_parser import VendorParser


class AppBioQuantStudioDesignandanalysisParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = DesignQuantstudioContents(named_file_contents.contents)
        data = Data.create(contents)
        return self._get_model(data, named_file_contents.original_file_name)

    def _get_model(self, data: Data, file_name: str) -> Model:
        return Model(
            qPCR_aggregate_document=QPCRAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    product_manufacturer="ThermoFisher Scientific",
                    device_identifier=data.header.device_identifier,
                    model_number=data.header.model_number,
                    device_serial_number=data.header.device_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier="localhost",
                    file_name=file_name,
                    UNC_path="",  # unknown
                    software_name=data.header.software_name,
                    software_version=data.header.software_version,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                qPCR_document=[
                    QPCRDocumentItem(
                        analyst=data.header.analyst,
                        measurement_aggregate_document=self.get_measurement_aggregate_document(
                            data, well
                        ),
                    )
                    for well in data.wells
                ],
            )
        )

    def get_measurement_aggregate_document(
        self, data: Data, well: Well
    ) -> MeasurementAggregateDocument:
        return MeasurementAggregateDocument(
            experimental_data_identifier=data.header.experimental_data_identifier,
            container_type=ContainerType.qPCR_reaction_block,
            plate_well_count=TQuantityValueNumber(value=data.header.plate_well_count),
            measurement_document=[
                self.get_measurement_document_item(data, well, well_item)
                for well_item in well.items.values()
            ],
        )

    def get_measurement_document_item(
        self, data: Data, well: Well, well_item: WellItem
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=well_item.uuid,
            measurement_time=self._get_date_time(data.header.measurement_time),
            target_DNA_description=well_item.target_dna_description,
            sample_document=self.get_sample_document(data, well_item),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    self.get_device_control_document_item(data, well_item),
                ],
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    self.get_processed_data_document(well_item),
                ]
            ),
            reporter_dye_data_cube=self.get_reporter_dye_data_cube(well, well_item),
            passive_reference_dye_data_cube=self.get_passive_reference_dye_data_cube(
                data, well
            ),
            melting_curve_data_cube=self.get_melting_curve_data_cube(well_item),
        )

    def get_sample_document(self, data: Data, well_item: WellItem) -> SampleDocument:
        return SampleDocument(
            sample_identifier=well_item.sample_identifier,
            sample_role_type=well_item.sample_role_type,
            well_location_identifier=well_item.well_location_identifier,
            well_plate_identifier=data.header.barcode,
        )

    def get_device_control_document_item(
        self, data: Data, well_item: WellItem
    ) -> DeviceControlDocumentItem:
        return DeviceControlDocumentItem(
            device_type="qPCR",
            measurement_method_identifier=data.header.measurement_method_identifier,
            total_cycle_number_setting=TQuantityValueNumber(
                value=well_item.amplification_data.total_cycle_number_setting,
            ),
            PCR_detection_chemistry=data.header.pcr_detection_chemistry,
            reporter_dye_setting=well_item.reporter_dye_setting,
            quencher_dye_setting=well_item.quencher_dye_setting,
            passive_reference_dye_setting=data.header.passive_reference_dye_setting,
        )

    def get_processed_data_document(
        self, well_item: WellItem
    ) -> ProcessedDataDocumentItem:
        normalized_reporter_result = (
            None
            if well_item.result.normalized_reporter_result is None
            else TQuantityValueUnitless(
                value=well_item.result.normalized_reporter_result
            )
        )

        baseline_corrected_reporter_result = (
            None
            if well_item.result.baseline_corrected_reporter_result is None
            else TQuantityValueUnitless(
                value=well_item.result.baseline_corrected_reporter_result
            )
        )

        return ProcessedDataDocumentItem(
            data_processing_document=self.get_data_processing_document(well_item),
            cycle_threshold_result=TNullableQuantityValueUnitless(
                value=well_item.result.cycle_threshold_result,
            ),
            normalized_reporter_result=normalized_reporter_result,
            baseline_corrected_reporter_result=baseline_corrected_reporter_result,
            genotyping_determination_result=well_item.result.genotyping_determination_result,
            normalized_reporter_data_cube=NormalizedReporterDataCube(
                label="normalized reporter",
                cube_structure=TDatacubeStructure(
                    dimensions=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.integer,
                            concept="cycle count",
                            unit="#",
                        ),
                    ],
                    measures=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.double,
                            concept="normalized report result",
                            unit="(unitless)",
                        ),
                    ],
                ),
                data=TDatacubeData(
                    dimensions=[well_item.amplification_data.cycle],
                    measures=[well_item.amplification_data.rn],
                ),
            ),
            baseline_corrected_reporter_data_cube=BaselineCorrectedReporterDataCube(
                label="baseline corrected reporter",
                cube_structure=TDatacubeStructure(
                    dimensions=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.integer,
                            concept="cycle count",
                            unit="#",
                        ),
                    ],
                    measures=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.double,
                            concept="baseline corrected reporter result",
                            unit="(unitless)",
                        ),
                    ],
                ),
                data=TDatacubeData(
                    dimensions=[well_item.amplification_data.cycle],
                    measures=[well_item.amplification_data.delta_rn],
                ),
            ),
        )

    def get_data_processing_document(
        self, well_item: WellItem
    ) -> DataProcessingDocument:
        baseline_determination_start_cycle_setting = (
            None
            if well_item.result.baseline_determination_start_cycle_setting is None
            else TQuantityValueNumber(
                value=well_item.result.baseline_determination_start_cycle_setting
            )
        )

        baseline_determination_end_cycle_setting = (
            None
            if well_item.result.baseline_determination_end_cycle_setting is None
            else TQuantityValueNumber(
                value=well_item.result.baseline_determination_end_cycle_setting
            )
        )

        genotyping_determination_method_setting = (
            None
            if well_item.result.genotyping_determination_method_setting is None
            else TQuantityValueUnitless(
                value=well_item.result.genotyping_determination_method_setting
            )
        )

        return DataProcessingDocument(
            automatic_cycle_threshold_enabled_setting=well_item.result.automatic_cycle_threshold_enabled_setting,
            cycle_threshold_value_setting=TQuantityValueUnitless(
                value=well_item.result.cycle_threshold_value_setting,
            ),
            automatic_baseline_determination_enabled_setting=well_item.result.automatic_baseline_determination_enabled_setting,
            baseline_determination_start_cycle_setting=baseline_determination_start_cycle_setting,
            baseline_determination_end_cycle_setting=baseline_determination_end_cycle_setting,
            genotyping_determination_method_setting=genotyping_determination_method_setting,
        )

    def get_reporter_dye_data_cube(
        self, well: Well, well_item: WellItem
    ) -> Optional[ReporterDyeDataCube]:
        if well.multicomponent_data is None or well_item.reporter_dye_setting is None:
            return None

        return ReporterDyeDataCube(
            label="reporter dye",
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.integer,
                        concept="cycle count",
                        unit="#",
                    ),
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="reporter dye fluorescence",
                        unit="RFU",
                    ),
                ],
            ),
            data=TDatacubeData(
                dimensions=[well.multicomponent_data.cycle],
                measures=[
                    well.multicomponent_data.get_column(well_item.reporter_dye_setting)
                ],
            ),
        )

    def get_passive_reference_dye_data_cube(
        self, data: Data, well: Well
    ) -> Optional[PassiveReferenceDyeDataCube]:
        if (
            well.multicomponent_data is None
            or data.header.passive_reference_dye_setting is None
        ):
            return None

        return PassiveReferenceDyeDataCube(
            label="passive reference dye",
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.integer,
                        concept="cycle count",
                        unit="#",
                    ),
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="passive reference dye fluorescence",
                        unit="RFU",
                    ),
                ],
            ),
            data=TDatacubeData(
                dimensions=[well.multicomponent_data.cycle],
                measures=[
                    well.multicomponent_data.get_column(
                        data.header.passive_reference_dye_setting
                    )
                ],
            ),
        )

    def get_melting_curve_data_cube(
        self, well_item: WellItem
    ) -> Optional[MeltingCurveDataCube]:
        if well_item.melt_curve_data is None:
            return None

        return MeltingCurveDataCube(
            label="melt curve",
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="temperature",
                        unit="degC",
                    ),
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="fluorescence",
                        unit="RFU",
                    ),
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="derivative",
                        unit="unitless",
                    ),
                ],
            ),
            data=TDatacubeData(
                dimensions=[well_item.melt_curve_data.temperature],
                measures=[
                    well_item.melt_curve_data.fluorescence,
                    well_item.melt_curve_data.derivative,
                ],
            ),
        )



from typing import Any, Union

from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    ContainerType as FluorescenceContainerType,
    MeasurementAggregateDocument as FluorescenceMeasurementAggregateDocument,
    Model as FluorescenceModel,
)
from allotropy.allotrope.models.luminescence_benchling_2023_09_luminescence import (
    ContainerType as LuminescenceContainerType,
    MeasurementAggregateDocument as LuminescenceMeasurementAggregateDocument,
    Model as LuminescenceModel,
)
from allotropy.allotrope.models.shared.definitions.custom import TQuantityValueNumber
from allotropy.allotrope.models.ultraviolet_absorbance_benchling_2023_09_ultraviolet_absorbance import (
    ContainerType as AbsorbanceContainerType,
    MeasurementAggregateDocument as AbsorbanceMeasurementAggregateDocument,
    Model as AbsorbanceModel,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import Data
from allotropy.parsers.agilent_gen5.constants import ReadMode
from allotropy.parsers.agilent_gen5.plate_data import PlateData
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.lines_reader import read_to_lines
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5Parser(VendorParser):
    def _create_model(
        self, plate_data: PlateData
    ) -> Union[AbsorbanceModel, FluorescenceModel, LuminescenceModel]:
        measurement_docs = plate_data.results.measurement_docs
        if plate_data.read_data.read_mode == ReadMode.ABSORBANCE:
            return AbsorbanceModel(
                measurement_aggregate_document=AbsorbanceMeasurementAggregateDocument(
                    measurement_identifier=random_uuid_str(),
                    measurement_time=self._get_date_time(
                        plate_data.header_data.datetime
                    ),
                    analytical_method_identifier=plate_data.file_paths.protocol_file_path,
                    experimental_data_identifier=plate_data.file_paths.experiment_file_path,
                    container_type=AbsorbanceContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(
                        len(plate_data.results.wells)
                    ),
                    measurement_document=measurement_docs,
                )
            )
        elif plate_data.read_data.read_mode == ReadMode.FLUORESCENCE:
            return FluorescenceModel(
                measurement_aggregate_document=FluorescenceMeasurementAggregateDocument(
                    measurement_identifier=random_uuid_str(),
                    measurement_time=self._get_date_time(
                        plate_data.header_data.datetime
                    ),
                    analytical_method_identifier=plate_data.file_paths.protocol_file_path,
                    experimental_data_identifier=plate_data.file_paths.experiment_file_path,
                    container_type=FluorescenceContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(
                        len(plate_data.results.wells)
                    ),
                    measurement_document=measurement_docs,
                )
            )
        elif plate_data.read_data.read_mode == ReadMode.LUMINESCENCE:
            return LuminescenceModel(
                measurement_aggregate_document=LuminescenceMeasurementAggregateDocument(
                    measurement_identifier=random_uuid_str(),
                    measurement_time=self._get_date_time(
                        plate_data.header_data.datetime
                    ),
                    analytical_method_identifier=plate_data.file_paths.protocol_file_path,
                    experimental_data_identifier=plate_data.file_paths.experiment_file_path,
                    container_type=LuminescenceContainerType.well_plate,
                    plate_well_count=TQuantityValueNumber(
                        len(plate_data.results.wells)
                    ),
                    measurement_document=measurement_docs,
                )
            )

        msg = msg_for_error_on_unrecognized_value(
            "read mode", plate_data.read_data.read_mode, ReadMode._member_names_
        )
        raise AllotropeConversionError(msg)

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Any:
        lines = read_to_lines(named_file_contents)
        section_lines_reader = SectionLinesReader(lines)
        data = Data.create(section_lines_reader)

        return self._create_model(data.plate_data)



import re

from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.revvity_kaleido.kaleido_structure import (
    AnalysisResult,
    BackgroundInfo,
    Data,
    MeasurementElement,
    MeasurementInfo,
    Measurements,
    Platemap,
    Results,
    SCAN_POSITION_CONVERSION,
)
from allotropy.parsers.utils.values import assert_not_none, try_float_or_none


def create_background_info(reader: CsvReader) -> BackgroundInfo:
    line = assert_not_none(
        reader.drop_until_inclusive("^Results for.(.+) 1"),
        msg="Unable to find background information.",
    )

    experiment_type = assert_not_none(
        re.match("^Results for.(.+) 1", line),
        msg="Unable to find experiment type from background information section.",
    ).group(1)

    return BackgroundInfo(experiment_type)


def create_results(reader: CsvReader) -> Results:
    barcode_line = assert_not_none(
        reader.drop_until_inclusive("^Barcode:(.+),"),
        msg="Unable to find background information.",
    )

    raw_barcode, *_ = barcode_line.split(",")
    barcode = raw_barcode.removeprefix("Barcode:")

    results = assert_not_none(
        reader.pop_csv_block_as_df(header=0, index_col=0),
        msg="Unable to find results table.",
    )

    for column in results:
        if str(column).startswith("Unnamed"):
            results = results.drop(columns=column)

    return Results(
        barcode=barcode,
        data={
            f"{row}{col}": values[col]
            for row, values in results.iterrows()
            for col in results.columns
        },
    )


def create_analysis_results(reader: CsvReader) -> list[AnalysisResult]:
    section_title = assert_not_none(
        reader.drop_until("^Results for|^Measurement Basic Information"),
        msg="Unable to find Analysis Result or Measurement Basic Information section.",
    )

    if section_title.startswith("Measurement Basic Information"):
        return []

    reader.drop_until("^Barcode")

    analysis_results = []
    while reader.match("^Barcode"):
        if analysis_result := AnalysisResult.create(reader):
            analysis_results.append(analysis_result)

    return analysis_results


def create_measurement_info(reader: CsvReader) -> MeasurementInfo:
    assert_not_none(
        reader.drop_until_inclusive("^Measurement Basic Information"),
        msg="Unable to find Measurement Basic Information section.",
    )

    elements = {}
    for raw_line in reader.pop_until("^Plate Type"):
        if raw_line == "":
            continue

        key, _, value, *_ = raw_line.split(",")
        elements[key.rstrip(":")] = value

    return MeasurementInfo.create(elements)


def create_platemap(reader: CsvReader) -> Platemap:
    assert_not_none(
        reader.drop_until_inclusive("^Platemap"),
        msg="Unable to find Platemap section.",
    )

    assert_not_none(
        reader.pop_if_match("^Plate"),
        msg="Unable to find platemap start indicator.",
    )

    data = assert_not_none(
        reader.pop_csv_block_as_df(header=0, index_col=0),
        msg="Unable to find platemap information.",
    )

    return Platemap(
        data={
            f"{row}{col}": values[col]
            for row, values in data.iterrows()
            for col in data.columns
        }
    )


def create_measurements(reader: CsvReader) -> Measurements:
    assert_not_none(
        reader.drop_until_inclusive("^Measurements"),
        msg="Unable to find Measurements section.",
    )

    elements = []
    for raw_line in reader.pop_until("^Analysis"):
        if raw_line == "":
            continue

        key, _, _, _, value, *_ = raw_line.split(",")
        elements.append(
            MeasurementElement(title=key.rstrip(":"), value=value),
        )

    scan_position_element = Measurements.try_element_or_none(
        elements, "Excitation / Emission"
    )
    excitation_wavelength_element = Measurements.try_element_or_none(
        elements, "Excitation wavelength [nm]"
    )

    return Measurements(
        channels=Measurements.create_channels(elements),
        number_of_averages=Measurements.get_element_float_value_or_none(
            elements, "Number of flashes"
        ),
        detector_distance=Measurements.get_element_float_value_or_none(
            elements, "Distance between Plate and Detector [mm]"
        ),
        scan_position=(
            None
            if scan_position_element is None
            else assert_not_none(
                SCAN_POSITION_CONVERSION.get(scan_position_element.value),
                msg=f"'{scan_position_element.value}' is not a valid scan position, expected TOP or BOTTOM.",
            )
        ),
        emission_wavelength=Measurements.get_element_float_value_or_none(
            elements, "Emission wavelength [nm]"
        ),
        excitation_wavelength=(
            None
            if excitation_wavelength_element is None
            else try_float_or_none(
                excitation_wavelength_element.value.removesuffix("nm")
            )
        ),
        focus_height=Measurements.get_element_float_value_or_none(
            elements, "Focus Height [µm]"
        ),
    )


def create_data_v2(version: str, reader: CsvReader) -> Data:
    background_info = create_background_info(reader)
    results = create_results(reader)
    analysis_results = create_analysis_results(reader)
    measurement_info = create_measurement_info(reader)

    assert_not_none(
        reader.drop_until_inclusive("^Plate Type"),
        msg="Unable to find Plate Type section.",
    )
    reader.drop_until("^Platemap")

    platemap = create_platemap(reader)
    measurements = create_measurements(reader)

    return Data(
        version,
        background_info,
        results,
        analysis_results,
        measurement_info,
        platemap,
        measurements,
    )



from enum import Enum
from typing import Optional

MULTIPLATE_FILE_ERROR = "Only a single plate per file can be processed at this time. Please refer to Gen5 documentation for how to generate single plate exports from multi-plate experiments"
NO_PLATE_DATA_ERROR = "No plate data found in file."
UNSUPORTED_READ_TYPE_ERROR = "Only Endpoint measurements can be processed at this time."

EMISSION_KEY = "Emission"
EXCITATION_KEY = "Excitation"
GAIN_KEY = "Gain"
OPTICS_KEY = "Optics"
MEASUREMENTS_DATA_POINT_KEY = "Measurements/Data Point"
MIRROR_KEY = "Mirror"
READ_HEIGHT_KEY = "Read Height"
READ_SPEED_KEY = "Read Speed"


class ReadMode(str, Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"
    LUMINESCENCE = "Luminescence"


class ReadType(str, Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    AREASCAN = "Area Scan"
    SPECTRAL = "Spectral"


READTYPE_TO_DIMENSIONS: dict[ReadType, list[tuple[str, str, Optional[str]]]] = {
    ReadType.ENDPOINT: [("int", "wavelength", "nm")],
    ReadType.KINETIC: [("double", "time", "s")],
    ReadType.AREASCAN: [
        ("int", "x", None),
        ("int", "y", None),
    ],
    ReadType.SPECTRAL: [("int", "wavelength", "nm")],
}



from __future__ import annotations

from dataclasses import dataclass

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.agilent_gen5.constants import MULTIPLATE_FILE_ERROR
from allotropy.parsers.agilent_gen5.plate_data import PlateData
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader


@dataclass(frozen=True)
class Data:
    plate_data: PlateData

    @staticmethod
    def create(section_lines_reader: SectionLinesReader) -> Data:
        plates = list(section_lines_reader.iter_sections("^Software Version"))

        if not plates:
            msg = "No plate data found in file."
            raise AllotropeConversionError(msg)

        if len(plates) > 1:
            raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

        return Data(PlateData.create(plates[0]))



import re

from allotropy.allotrope.models.shared.definitions.units import UNITLESS

WAVELENGTH_COLUMNS_RE = re.compile(r"^A\d{3}$")
NO_WAVELENGTH_COLUMN_ERROR_MSG = (
    "The file is required to include an absorbance measurement column (e.g. A280)"
)
INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG = (
    msg
) = "The wavelength_column should follow the pattern A### (ex. A260)."
NO_DATE_OR_TIME_ERROR_MSG = (
    "The file is required to include both 'Date' and 'Time' columns."
)
NO_MEASUREMENT_IN_PLATE_ERROR_MSG = (
    "The plate data does not contain absorbance measurement for {}."
)

CALCULATED_DATA_LOOKUP = {
    "A260": [
        {
            "column": "A260 Concentration (ng/ul)",
            "name": "Concentration",
            "feature": "absorbance",
            "unit": "ng/mL",
        },
        {
            "column": "Concentration (ng/ul)",
            "name": "Concentration",
            "feature": "absorbance",
            "unit": "ng/mL",
        },
        {
            "column": "Background (A260)",
            "name": "Background (A260)",
            "feature": "absorbance",
            "unit": "mAU",
        },
        {
            "column": "A260/A230",
            "name": "A260/A230",
            "feature": "absorbance",
            "unit": UNITLESS,
        },
        {
            "column": "A260/A280",
            "name": "A260/A280",
            "feature": "absorbance",
            "unit": UNITLESS,
        },
    ],
    "A280": [
        {
            "column": "Concentration (mg/ml)",
            "name": "Concentration",
            "feature": "absorbance",
            "unit": "mg/mL",
        },
        {
            "column": "Background (A280)",
            "name": "Background (A280)",
            "feature": "absorbance",
            "unit": "mAU",
        },
        {
            "column": "A260/A280",
            "name": "A260/A280",
            "feature": "absorbance",
            "unit": UNITLESS,
        },
    ],
}



#!/usr/bin/env python3
from pathlib import Path

import click

from allotropy.allotrope.schema_parser.generate_schemas import generate_schemas

ROOT_DIR = Path(__file__).parent.parent


@click.command()
@click.option("-r", "--regex", help="Regex to determine which schemas to generate.")
def _generate_schemas(regex: str) -> None:
    generate_schemas(ROOT_DIR, schema_regex=regex)


if __name__ == "__main__":
    _generate_schemas()



#!/usr/bin/env python

"""Convert Weyland-Yutani data to dataframe."""

import argparse
import sys
from typing import Any

import pandas as pd

from allotropy.parser_factory import Vendor
from allotropy.to_allotrope import allotrope_model_from_io
from allotropy.types import IOType


def main() -> None:
    """Main driver."""
    args = parse_args()
    for filename in args.filenames:
        with open(filename) as reader:
            handle(filename, reader)


def extract_well_data(model: Any) -> list[Any]:
    """Get (col, row, reading) from data."""

    def _split(loc: Any) -> tuple:  # type: ignore[type-arg]
        return (loc[0], int(loc[1:]))

    return [
        (*_split(doc.sample_document.well_location_identifier), 0.0)
        for doc in model.measurement_aggregate_document.measurement_document
    ]


def handle(filename: str, reader: IOType) -> None:
    """Extract and show data from a single model."""
    try:
        model = allotrope_model_from_io(reader, filename, Vendor.EXAMPLE_WEYLAND_YUTANI)
        well_data = extract_well_data(model)
        df = pd.DataFrame(well_data, columns=["col", "row", "reading"])
        print(df)  # noqa: T201
    except Exception as exc:
        print(f"Unable to read {filename}: {exc}", file=sys.stderr)  # noqa: T201


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("filenames", type=str, nargs="+", help="input file(s)")
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.types import IOType


class NucleoviewReader:
    @classmethod
    def read(cls, contents: IOType) -> pd.DataFrame:
        df = read_csv(
            contents,
            skipfooter=1,
            sep=";",
            usecols=[0, 1],
            skipinitialspace=True,
            index_col=False,
        ).dropna(axis=0, how="all")

        # add a common index to all rows for our group by and pivot
        df["group_by"] = ["Group1"] * len(df)

        # pivot to wide format
        raw_data = df.pivot(
            index="group_by", columns=df.columns[0], values=df.columns[1]
        )

        # give timezone offset a positive symbol if not present
        if "Time zone offset" in raw_data.columns and "Date time" in raw_data.columns:
            raw_data.loc[
                ~raw_data["Time zone offset"].str.startswith("-"), "Time zone offset"
            ] = ("+" + raw_data["Time zone offset"])

            # combine date time and UTC offset
            raw_data["datetime"] = pd.to_datetime(
                raw_data["Date time"] + raw_data["Time zone offset"]
            )
        raw_data["Sample ID"] = raw_data["Image"].str.split("-", n=3).str[3]
        raw_data = raw_data.rename(
            {"Estimated cell diameter [um]": "Estimated cell diameter (um)"},
            axis=1,
        )

        return raw_data



# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Union

from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.agilent_gen5.absorbance_data_point import AbsorbanceDataPoint
from allotropy.parsers.agilent_gen5.constants import (
    EMISSION_KEY,
    EXCITATION_KEY,
    GAIN_KEY,
    MEASUREMENTS_DATA_POINT_KEY,
    MIRROR_KEY,
    OPTICS_KEY,
    READ_HEIGHT_KEY,
    READ_SPEED_KEY,
    ReadMode,
    ReadType,
    UNSUPORTED_READ_TYPE_ERROR,
)
from allotropy.parsers.agilent_gen5.data_point import DataPoint
from allotropy.parsers.agilent_gen5.fluorescence_data_point import FluorescenceDataPoint
from allotropy.parsers.agilent_gen5.luminescence_data_point import LuminescenceDataPoint
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.values import assert_not_none, try_float, try_float_or_none

METADATA_PREFIXES = frozenset(
    {
        "Plate Number",
        "Date",
        "Time",
        "Reader Type:",
        "Reader Serial Number:",
        "Reading Type",
    }
)
GEN5_DATETIME_FORMAT = "%m/%d/%Y %I:%M:%S %p"


def try_float_or_value(value: str) -> Union[str, float]:
    try:
        return float(value)
    except ValueError:
        return value


def read_data_section(reader: LinesReader) -> str:
    data_section = "\n".join(
        [
            reader.pop() or "",
            *reader.pop_until_empty(),
        ]
    )
    reader.drop_empty()
    return data_section


def hhmmss_to_sec(hhmmss: str) -> int:
    # convert to seconds, as specified by Allotrope
    hours, minutes, seconds = tuple(int(num) for num in hhmmss.split(":"))
    return (3600 * hours) + (60 * minutes) + seconds


@dataclass(frozen=True)
class FilePaths:
    experiment_file_path: str
    protocol_file_path: str

    @staticmethod
    def create(reader: LinesReader) -> FilePaths:
        assert_not_none(
            reader.drop_until("^Experiment File Path"),
            "Experiment File Path",
        )
        file_paths = reader.pop_until_empty()
        return FilePaths(
            experiment_file_path=f"{next(file_paths)}\t".split("\t")[1],
            protocol_file_path=f"{next(file_paths)}\t".split("\t")[1],
        )


@dataclass(frozen=True)
class HeaderData:
    datetime: str
    well_plate_identifier: str
    model_number: str
    equipment_serial_number: str

    @classmethod
    def create(cls, reader: LinesReader) -> HeaderData:
        assert_not_none(reader.drop_until("^Plate Number"), "Plate Number")
        metadata_dict = cls._parse_metadata(reader)
        datetime_ = cls._parse_datetime(metadata_dict["Date"], metadata_dict["Time"])

        return HeaderData(
            datetime=datetime_,
            well_plate_identifier=metadata_dict["Plate Number"],
            model_number=metadata_dict["Reader Type:"],
            equipment_serial_number=metadata_dict["Reader Serial Number:"],
        )

    @classmethod
    def _parse_metadata(cls, reader: LinesReader) -> dict:
        metadata_dict: dict = {}
        for metadata_line in reader.pop_until_empty():
            line_split = metadata_line.split("\t")
            if line_split[0] not in METADATA_PREFIXES:
                msg = msg_for_error_on_unrecognized_value(
                    "metadata key", line_split[0], METADATA_PREFIXES
                )
                raise AllotropeConversionError(msg)
            metadata_dict[line_split[0]] = line_split[1]
        return metadata_dict

    @classmethod
    def _parse_datetime(cls, date_: str, time_: str) -> str:
        return f"{date_} {time_}"


@dataclass(frozen=True)
class ReadData:
    read_mode: ReadMode
    read_type: ReadType
    read_names: list[str]
    wavelengths: list[float]
    step_label: Optional[str]
    number_of_averages: Optional[float]
    emissions: Optional[list[str]]
    optics: Optional[list[str]]
    gains: Optional[list[float]]
    detector_distance: Optional[float]
    detector_carriage_speed: Optional[str]
    excitations: Optional[list[str]]
    wavelength_filter_cut_offs: Optional[list[float]]
    scan_positions: Optional[list[str]]

    @classmethod
    def create(cls, reader: LinesReader) -> ReadData:
        assert_not_none(reader.drop_until("^Procedure Details"), "Procedure Details")
        reader.pop()
        reader.drop_empty()
        procedure_details = read_data_section(reader)

        read_type = cls.get_read_type(procedure_details)
        if read_type != ReadType.ENDPOINT:
            raise AllotropeConversionError(UNSUPORTED_READ_TYPE_ERROR)

        read_mode = cls.get_read_mode(procedure_details)
        read_names = []
        procedure_chunks = cls._get_procedure_chunks(procedure_details)

        for procedure_chunk in procedure_chunks:
            read_names.extend(cls._parse_procedure_chunk(procedure_chunk, read_mode))

        device_control_data = cls._get_device_control_data(procedure_details, read_mode)

        wavelengths = [
            try_float(wavelength, "Wavelength")
            for wavelength in device_control_data.get("Wavelengths", [])
        ]
        gains = [
            try_float(gain, "Gain") for gain in device_control_data.get(GAIN_KEY, [])
        ]
        number_of_averages = device_control_data.get(MEASUREMENTS_DATA_POINT_KEY)
        read_height = device_control_data.get(READ_HEIGHT_KEY, "")

        mirrors = device_control_data.get(MIRROR_KEY, [])
        optics = device_control_data.get(OPTICS_KEY, [])
        scan_positions = []
        wavelength_filter_cut_offs = []
        if mirrors and read_mode == ReadMode.FLUORESCENCE:
            for mirror in mirrors:
                position, cutoff, *_ = mirror.split(" ")
                scan_positions.append(position)
                wavelength_filter_cut_offs.append(try_float(cutoff, "Cutoff"))
        elif optics and read_mode == ReadMode.FLUORESCENCE:
            scan_positions = optics

        return ReadData(
            read_mode=read_mode,
            # TODO: Remove read_type
            read_type=read_type,
            read_names=read_names,
            step_label=device_control_data.get("Step Label"),
            # Absorbance attributes
            wavelengths=wavelengths,
            detector_carriage_speed=device_control_data.get(READ_SPEED_KEY),
            number_of_averages=try_float_or_none(number_of_averages),
            # Luminescence attributes
            emissions=device_control_data.get(EMISSION_KEY),
            optics=device_control_data.get(OPTICS_KEY),
            gains=gains,
            detector_distance=(try_float_or_none(read_height.split(" ")[0])),
            # Fluorescence attributes
            excitations=device_control_data.get(EXCITATION_KEY),
            wavelength_filter_cut_offs=wavelength_filter_cut_offs,
            scan_positions=scan_positions,
        )

    @property
    def data_point_cls(self) -> type[DataPoint]:
        if self.read_mode == ReadMode.ABSORBANCE:
            return AbsorbanceDataPoint
        elif self.read_mode == ReadMode.FLUORESCENCE:
            return FluorescenceDataPoint
        elif self.read_mode == ReadMode.LUMINESCENCE:
            return LuminescenceDataPoint

        msg = msg_for_error_on_unrecognized_value(
            "read mode", self.read_mode, ReadMode._member_names_
        )
        raise AllotropeConversionError(msg)

    @staticmethod
    def get_read_mode(procedure_details: str) -> ReadMode:
        if ReadMode.ABSORBANCE.value in procedure_details:
            return ReadMode.ABSORBANCE
        elif ReadMode.FLUORESCENCE.value in procedure_details:
            return ReadMode.FLUORESCENCE
        elif ReadMode.LUMINESCENCE.value in procedure_details:
            return ReadMode.LUMINESCENCE

        msg = f"Read mode not found; expected to find one of {sorted(ReadMode._member_names_)}."
        raise AllotropeConversionError(msg)

    @staticmethod
    def get_read_type(procedure_details: str) -> ReadType:
        if ReadType.KINETIC.value in procedure_details:
            return ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL

        # check for this last, because other modes still contain the word "Endpoint"
        return ReadType.ENDPOINT

    @classmethod
    def _get_device_control_data(
        cls, procedure_details: str, read_mode: ReadMode
    ) -> dict:
        list_keys = frozenset(
            {
                EMISSION_KEY,
                EXCITATION_KEY,
                OPTICS_KEY,
                GAIN_KEY,
                MIRROR_KEY,
                "Wavelengths",
            }
        )
        read_data_dict: dict = {label: [] for label in list_keys}
        read_lines: list[str] = procedure_details.splitlines()
        datum_len = 2

        for line in read_lines:
            strp_line = str(line.strip())
            if strp_line.startswith("Read\t"):
                read_data_dict["Step Label"] = cls._get_step_label(line, read_mode)
                continue

            elif strp_line.startswith("Wavelengths"):
                wavelengths = strp_line.split(":  ")
                read_data_dict["Wavelengths"].extend(wavelengths[1].split(", "))
                continue

            elif strp_line.startswith("Pathlength Correction"):
                corrections = strp_line.split(": ")
                read_data_dict["Wavelengths"].extend(corrections[1].split("/"))
                continue

            line_data: list[str] = strp_line.split(",  ")
            for read_datum in line_data:
                splitted_datum = read_datum.split(": ")
                if len(splitted_datum) != datum_len:
                    continue
                if splitted_datum[0] in list_keys:
                    read_data_dict[splitted_datum[0]].append(splitted_datum[1])
                else:
                    read_data_dict[splitted_datum[0]] = splitted_datum[1]

        return read_data_dict

    @staticmethod
    def _get_procedure_chunks(procedure_details: str) -> list[list[str]]:
        procedure_chunks = []
        current_chunk: list[str] = []
        procedure_lines = procedure_details.splitlines()
        for procedure_line in procedure_lines:
            if procedure_line[0] != "\t":
                procedure_chunks.append(current_chunk)
                current_chunk = []
            current_chunk.append(procedure_line)
        del procedure_chunks[0]
        procedure_chunks.append(current_chunk)

        return procedure_chunks

    @classmethod
    def _get_step_label(cls, read_line: str, read_mode: str) -> Optional[str]:
        read_data_len = 2
        split_line = read_line.split("\t")
        if len(split_line) != read_data_len:
            msg = (
                f"Expected the Read data line {split_line} to contain exactly 2 values."
            )
            raise AllotropeConversionError(msg)
        if split_line[1] != f"{read_mode.title()} Endpoint":
            return split_line[1]

        return None

    @staticmethod
    def _parse_procedure_chunk(
        procedure_chunk: list[str],
        read_mode: ReadMode,
    ) -> list[str]:
        # TODO: remove when using wavelenghts and bandwiths along with stepLabels

        # if no user-defined name is specified for protocols,
        # e.g. it just says "Absorbance Endpoint",
        # Gen5 defaults to using the wavelength as the name
        read_names = []
        use_wavelength_names = False
        read_line_length = 2
        wavelength_line_length = 2

        for line in procedure_chunk:
            split_line = line.strip().split("\t")
            if split_line[0] == "Read":
                if len(split_line) != read_line_length:
                    msg = f"Expected the Read data line {split_line} to contain exactly {read_line_length} values."
                    raise AllotropeConversionError(msg)
                if split_line[-1] == f"{read_mode.title()} Endpoint":
                    use_wavelength_names = True
                else:
                    read_names.append(split_line[-1])
            elif split_line[0].startswith("Wavelengths"):
                if use_wavelength_names:
                    split_line_colon = split_line[0].split(":  ")
                    if len(split_line_colon) != wavelength_line_length:
                        msg = f"Expected the Wavelengths data line {split_line} to contain exactly {wavelength_line_length} values."
                        raise AllotropeConversionError(msg)
                    read_names.extend(split_line_colon[-1].split(", "))
        return read_names


@dataclass(frozen=True)
class LayoutData:
    layout: dict
    concentrations: dict

    @staticmethod
    def create_default() -> LayoutData:
        return LayoutData(
            layout={},
            concentrations={},
        )

    @staticmethod
    def create(layout_str: str) -> LayoutData:
        layout = {}
        concentrations = {}

        layout_lines = layout_str.splitlines()
        # first line is "Layout", second line is column numbers
        current_row = "A"
        for i in range(2, len(layout_lines)):
            split_line = layout_lines[i].split("\t")
            if split_line[0]:
                current_row = split_line[0]
            label = split_line[-1]
            for j in range(1, len(split_line) - 1):
                well_loc = f"{current_row}{j}"
                if label == "Well ID":
                    layout[well_loc] = split_line[j]
                elif label == "Conc/Dil" and split_line[j]:
                    concentrations[well_loc] = float(split_line[j])

        return LayoutData(
            layout=layout,
            concentrations=concentrations,
        )


@dataclass(frozen=True)
class ActualTemperature:
    value: Optional[float] = None

    @staticmethod
    def create_default() -> ActualTemperature:
        return ActualTemperature()

    @staticmethod
    def create(actual_temperature: str) -> ActualTemperature:
        if len(actual_temperature.split("\n")) != 1:
            msg = f"Expected the Temperature section '{actual_temperature}' to contain exactly 1 line."
            raise AllotropeConversionError(msg)

        return ActualTemperature(
            value=float(actual_temperature.strip().split("\t")[-1]),
        )


@dataclass(frozen=True)
class Results:
    measurements: defaultdict[str, list]
    processed_datas: defaultdict[str, list]
    wells: list
    measurement_docs: list

    @staticmethod
    def create() -> Results:
        return Results(
            measurements=defaultdict(list),
            processed_datas=defaultdict(list),
            wells=[],
            measurement_docs=[],
        )

    def parse_results(
        self,
        results: str,
        read_data: ReadData,
        header_data: HeaderData,
        layout_data: LayoutData,
        actual_temperature: ActualTemperature,
    ) -> None:
        result_lines = results.splitlines()
        if result_lines[0].strip() != "Results":
            msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
            raise AllotropeConversionError(msg)
        # result_lines[1] contains column numbers

        current_row = "A"
        for row_num in range(2, len(result_lines)):
            values = result_lines[row_num].split("\t")
            if values[0]:
                current_row = values[0]
            label = values[-1]  # last column gives information about the type of read
            for col_num in range(1, len(values) - 1):
                well_pos = f"{current_row}{col_num}"
                if well_pos not in self.wells:
                    self.wells.append(well_pos)
                well_value: Union[str, float] = try_float_or_value(values[col_num])
                if Results._is_processed_data_label(
                    label,
                    read_data.read_mode,
                    read_data.read_names,
                ):
                    self.processed_datas[well_pos].append([label, well_value])
                else:
                    label_only = label.split(":")[-1]
                    self.measurements[well_pos].append([label_only, well_value])

        for well_pos in self.wells:
            self.measurement_docs.append(
                read_data.data_point_cls(
                    read_type=read_data.read_type,
                    measurements=self.measurements[well_pos],
                    well_location=well_pos,
                    well_plate_identifier=header_data.well_plate_identifier,
                    sample_identifier=layout_data.layout.get(well_pos),
                    concentration=layout_data.concentrations.get(well_pos),
                    processed_data=self.processed_datas[well_pos],
                    temperature=actual_temperature.value,
                ).to_measurement_doc()
            )

    @staticmethod
    def _is_processed_data_label(
        label: str,
        read_mode: ReadMode,
        read_names: list,
    ) -> bool:
        if read_mode == ReadMode.LUMINESCENCE:
            return not any(
                (label.startswith(read_name) and label.split(":")[-1] == "Lum")
                for read_name in read_names
            )
        else:
            return not any(
                (label.startswith(read_name) and label.split(":")[-1][0].isdigit())
                for read_name in read_names
            )


@dataclass(frozen=True)
class PlateData:
    file_paths: FilePaths
    header_data: HeaderData
    read_data: ReadData
    results: Results

    @staticmethod
    def create(reader: LinesReader) -> PlateData:
        file_paths = FilePaths.create(reader)
        header_data = HeaderData.create(reader)
        read_data = ReadData.create(reader)
        layout_data = LayoutData.create_default()
        actual_temperature = ActualTemperature.create_default()
        results = Results.create()

        while reader.current_line_exists():
            data_section = read_data_section(reader)
            if data_section.startswith("Layout"):
                layout_data = LayoutData.create(data_section)
            elif data_section.startswith("Actual Temperature"):
                actual_temperature = ActualTemperature.create(data_section)
            elif data_section.startswith("Results"):
                results.parse_results(
                    data_section,
                    read_data,
                    header_data,
                    layout_data,
                    actual_temperature,
                )

        return PlateData(
            file_paths=file_paths,
            header_data=header_data,
            read_data=read_data,
            results=results,
        )



from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_calculated_documents import (
    build_quantity,
    iter_calculated_data_documents,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    AmplificationData,
    Data,
    Header,
    MeltCurveRawData,
    MulticomponentData,
    RawData,
    Result,
    WellList,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.values import try_str_from_series_or_none


def create_data(reader: LinesReader) -> Data:
    header = Header.create(reader)
    wells = WellList.create(reader, header.experiment_type)
    raw_data = RawData.create(reader)

    amp_data = AmplificationData.get_data(reader)
    multi_data = MulticomponentData.get_data(reader)
    results_data, results_metadata = Result.get_data(reader)
    melt_data = MeltCurveRawData.get_data(reader)
    for well in wells:
        if multi_data is not None:
            well.multicomponent_data = MulticomponentData.create(
                multi_data,
                well,
            )

        if melt_data is not None:
            well.melt_curve_raw_data = MeltCurveRawData.create(
                melt_data,
                well,
            )

        for well_item in well.items.values():
            well_item.amplification_data = AmplificationData.create(
                amp_data,
                well_item,
            )

            well_item.result = Result.create(
                results_data,
                well_item,
                header.experiment_type,
            )

        for well_item in well.items.values():
            if quantity := build_quantity(well_item):
                well.add_calculated_document(quantity)

    endogenous_control = (
        try_str_from_series_or_none(results_metadata, "Endogenous Control") or ""
    )
    reference_sample = (
        try_str_from_series_or_none(results_metadata, "Reference Sample") or ""
    )

    return Data(
        header,
        wells,
        raw_data,
        endogenous_control,
        reference_sample,
        list(
            iter_calculated_data_documents(
                wells,
                header.experiment_type,
                reference_sample,
                endogenous_control,
            )
        ),
    )



from allotropy.allotrope.models.cell_culture_analyzer_benchling_2023_09_cell_culture_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    SampleDocument,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.novabio_flex2.novabio_flex2_structure import Data, Sample
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser


class NovaBioFlexParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        return self._get_model(Data.create(named_file_contents))

    def _get_model(self, data: Data) -> Model:
        return Model(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_identifier=random_uuid_str(),
                data_processing_time=self._get_date_time(data.title.processing_time),
                analyst=data.sample_list.analyst,
                device_system_document=DeviceSystemDocument(
                    model_number="NovaBio Flex2",
                    device_identifier=data.title.device_identifier,
                ),
                measurement_document=self._get_measurement_document(data),
            ),
        )

    def _get_measurement_document(self, data: Data) -> list[MeasurementDocumentItem]:
        # get analytes and properties included in the output file
        return [
            self._get_measurement_from_sample(sample)
            for sample in data.sample_list.samples
        ]

    def _get_measurement_from_sample(self, sample: Sample) -> MeasurementDocumentItem:
        sample_measurement = MeasurementDocumentItem(
            sample_document=SampleDocument(
                sample_identifier=sample.identifier,
                sample_role_type=sample.role_type,
                batch_identifier=sample.batch_identifier,
            ),
            measurement_time=self._get_date_time(sample.measurement_time),
            analyte_aggregate_document=AnalyteAggregateDocument(
                analyte_document=[
                    AnalyteDocumentItem(
                        analyte_name=analyte.name,
                        molar_concentration=analyte.molar_concentration,
                    )
                    for analyte in sample.analytes
                ]
            ),
        )

        # properties are not included under the analyte aggregate document
        # and are added directly to the measurement document instead
        for name, property_ in sample.properties.items():
            setattr(sample_measurement, name, property_)

        return sample_measurement



import json
import os
from pathlib import Path
import re
from typing import Any

SCHEMAS_DIR = os.path.join(Path(__file__).parent, "schemas")


def add_definitions(schema: dict[str, Any]) -> dict[str, Any]:
    for file, section in [
        ("definitions", "defs"),
        ("units", "defs"),
        ("custom", "custom"),
    ]:
        existing = schema.get(f"${section}", {})
        with open(
            os.path.join(SCHEMAS_DIR, "shared", "definitions", f"{file}.json")
        ) as f:
            additional = json.load(f)
        additional.update(existing)
        schema[f"${section}"] = additional
    return schema


def get_schema(schema_relative_path: str) -> dict[str, Any]:
    with open(os.path.join(SCHEMAS_DIR, schema_relative_path)) as f:
        return add_definitions(json.load(f))


def get_schema_path_from_manifest(manifest: str) -> str:
    match = re.match(r"http://purl.allotrope.org/manifests/(.*)\.manifest", manifest)
    if not match:
        msg = f"No matching schema in repo for manifest: {manifest}"
        raise ValueError(msg)
    return f"{match.groups()[0]}.json"


def get_schema_from_manifest(manifest: str) -> dict[str, Any]:
    return get_schema(get_schema_path_from_manifest(manifest))


def get_schema_from_model(model: Any) -> dict[str, Any]:
    manifest = getattr(model, "manifest", getattr(model, "field_asm_manifest", None))
    if not manifest:
        msg = f"No 'manifest' or 'field_asm_manifest' found in model: {type(model)}"
        raise ValueError(msg)
    return get_schema_from_manifest(manifest)



from enum import Enum

from attr import dataclass


class AggregationType(str, Enum):
    INDIVIDUAL = "INDIVIDUAL"
    AVERAGE = "AVERAGE"
    POOLED = "POOLED"


class CalculatedDataSource(str, Enum):
    MEASUREMENT = "MEASUREMENT"
    CALCULATED_DATA = "CALCULATED_DATA"


AGGREGATION_LOOKUP = {
    "Average": AggregationType.AVERAGE,
    "Pooled": AggregationType.POOLED,
    None: AggregationType.INDIVIDUAL,
}


@dataclass(frozen=True)
class CalculatedDataItem:
    column: str
    name: str
    source: CalculatedDataSource
    source_feature: str
    unit: str


CALCULATED_DATA_REFERENCE: dict[AggregationType, list[CalculatedDataItem]] = {
    AggregationType.AVERAGE: [
        CalculatedDataItem(
            column="Total",
            name="Average Total",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count",
            unit="#",
        ),
        CalculatedDataItem(
            column="Conc. cp/uL",
            name="Average Concentration",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="number concentration",
            unit="#/μL",
        ),
        CalculatedDataItem(
            column="SD",
            name="SD Concentration",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="number concentration",
            unit="#/μL",
        ),
        CalculatedDataItem(
            column="CV%",
            name="CV",
            source=CalculatedDataSource.CALCULATED_DATA,
            source_feature="Average Concentration,SD Concentration",
            unit="%",
        ),
        CalculatedDataItem(
            column="Positives",
            name="Average Positives",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="positive partition count",
            unit="#",
        ),
    ],
    AggregationType.POOLED: [
        CalculatedDataItem(
            column="Total",
            name="Pooled Total",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="total partition count",
            unit="#",
        ),
        CalculatedDataItem(
            column="Conc. cp/uL",
            name="Pooled Concentration",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="number concentration",
            unit="#/μL",
        ),
        CalculatedDataItem(
            column="Positives",
            name="Pooled Positives",
            source=CalculatedDataSource.MEASUREMENT,
            source_feature="positive partition count",
            unit="#",
        ),
    ],
}



from dataclasses import dataclass

UNITLESS = "(unitless)"


@dataclass
class HasUnit:
    unit: str


@dataclass
class Centimeter(HasUnit):
    unit: str = "cm"


@dataclass
class Millimeter(HasUnit):
    unit: str = "mm"


@dataclass
class Nanometer(HasUnit):
    unit: str = "nm"


@dataclass
class Micrometer(HasUnit):
    unit: str = "μm"


@dataclass
class Hertz(HasUnit):
    unit: str = "Hz"


@dataclass
class CubicMillimeter(HasUnit):
    unit: str = "mm^3"


@dataclass
class Unitless(HasUnit):
    unit: str = UNITLESS


@dataclass
class SecondTime(HasUnit):
    unit: str = "s"


@dataclass
class Percent(HasUnit):
    unit: str = "%"


@dataclass
class Cell(HasUnit):
    unit: str = "cell"


@dataclass
class Microliter(HasUnit):
    unit: str = "μL"


@dataclass
class NumberPerMicroliter(HasUnit):
    unit: str = "#/μL"


@dataclass
class MilliSecond(HasUnit):
    unit: str = "ms"


@dataclass
class MillionCellsPerMilliliter(HasUnit):
    unit: str = "10^6 cells/mL"


@dataclass
class TODO(HasUnit):
    unit: str = "TODO"


@dataclass
class DegreeCelsius(HasUnit):
    unit: str = "degC"


@dataclass
class Number(HasUnit):
    unit: str = "#"


@dataclass
class NanogramPerMicroliter(HasUnit):
    unit: str = "ng/uL"


@dataclass
class MicrogramPerMicroliter(HasUnit):
    unit: str = "ug/uL"


@dataclass
class PicogramPerMilliliter(HasUnit):
    unit: str = "pg/mL"


@dataclass
class NanogramPerMilliliter(HasUnit):
    unit: str = "ng/mL"


@dataclass
class MicrogramPerMilliliter(HasUnit):
    unit: str = "ug/mL"


@dataclass
class MilligramPerMilliliter(HasUnit):
    unit: str = "mg/mL"


@dataclass
class GramPerLiter(HasUnit):
    unit: str = "g/L"


@dataclass
class UnitPerLiter(HasUnit):
    unit: str = "U/L"


@dataclass
class MillimeterOfMercury(HasUnit):
    unit: str = "mmHg"


@dataclass
class OpticalDensity(HasUnit):
    unit: str = "OD"


@dataclass
class PH(HasUnit):
    unit: str = "pH"


@dataclass
class MilliOsmolesPerKilogram(HasUnit):
    unit: str = "mosm/kg"


@dataclass
class MillimolePerLiter(HasUnit):
    unit: str = "mmol/L"


@dataclass
class MilliAbsorbanceUnit(HasUnit):
    unit: str = "mAU"


@dataclass
class RelativeFluorescenceUnit(HasUnit):
    unit: str = "RFU"


@dataclass
class RelativeLightUnit(HasUnit):
    unit: str = "RLU"


@dataclass
class SquareCentimetersPerGram(HasUnit):
    unit: str = "cm^2/g"



from collections.abc import Mapping
from typing import Optional, Union

import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicrogramPerMicroliter,
    TQuantityValueMicrogramPerMilliliter,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilligramPerMilliliter,
    TQuantityValueNanogramPerMicroliter,
    TQuantityValueNanogramPerMilliliter,
    TQuantityValueNanometer,
    TQuantityValuePicogramPerMilliliter,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    JsonFloat,
    TQuantityValue,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.models.spectrophotometry_benchling_2023_12_spectrophotometry import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SpectrophotometryAggregateDocument,
    SpectrophotometryDocumentItem,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_reader import (
    NanoDropEightReader,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none
from allotropy.parsers.vendor_parser import VendorParser

ConcentrationType = Union[
    TQuantityValueMicrogramPerMicroliter,
    TQuantityValueMicrogramPerMilliliter,
    TQuantityValueMilligramPerMilliliter,
    TQuantityValueNanogramPerMicroliter,
    TQuantityValueNanogramPerMilliliter,
    TQuantityValuePicogramPerMilliliter,
]
ConcentrationClassType = Union[
    type[TQuantityValueMicrogramPerMicroliter],
    type[TQuantityValueMicrogramPerMilliliter],
    type[TQuantityValueMilligramPerMilliliter],
    type[TQuantityValueNanogramPerMicroliter],
    type[TQuantityValueNanogramPerMilliliter],
    type[TQuantityValuePicogramPerMilliliter],
]

CONCENTRATION_UNIT_TO_TQUANTITY: Mapping[str, ConcentrationClassType] = {
    "ug/ul": TQuantityValueMicrogramPerMicroliter,
    "ug/ml": TQuantityValueMicrogramPerMilliliter,
    "mg/ml": TQuantityValueMilligramPerMilliliter,
    "ng/ul": TQuantityValueNanogramPerMicroliter,
    "ng/ml": TQuantityValueNanogramPerMilliliter,
    "pg/ul": TQuantityValuePicogramPerMilliliter,
}


def _get_str_or_none(data_frame: pd.DataFrame, row: int, column: str) -> Optional[str]:
    if column not in data_frame.columns:
        return None

    val = data_frame.iloc[row][column]
    if pd.isna(val):
        return None

    return str(val)


def _get_str(data_frame: pd.DataFrame, row: int, column: str) -> str:
    val = _get_str_or_none(data_frame=data_frame, row=row, column=column)

    assert_not_none(val)

    return str(val)


def _get_float(data_frame: pd.DataFrame, row: int, column: str) -> JsonFloat:
    try:
        return float(data_frame.iloc[row][column])
    except (ValueError, TypeError):
        return InvalidJsonFloat.NaN


def _get_concentration(
    conc: JsonFloat, unit: Optional[str]
) -> Optional[ConcentrationType]:
    if unit and unit in CONCENTRATION_UNIT_TO_TQUANTITY and isinstance(conc, float):
        cls = CONCENTRATION_UNIT_TO_TQUANTITY[unit]
        return cls(value=conc)

    return None


class NanodropEightParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        data = NanoDropEightReader.read(named_file_contents)
        data = self._add_measurement_uuids(data)
        return self._get_model(data, named_file_contents.original_file_name)

    def _get_model(self, data: pd.DataFrame, filename: str) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/spectrophotometry/BENCHLING/2023/12/spectrophotometry.manifest",
            spectrophotometry_aggregate_document=SpectrophotometryAggregateDocument(
                spectrophotometry_document=self._get_spectrophotometry_document(data),
                calculated_data_aggregate_document=CalculatedDataAggregateDocument(
                    calculated_data_document=self._get_calculated_data_document(data),
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                device_system_document=DeviceSystemDocument(
                    model_number="Nanodrop Eight",
                    device_identifier="Nanodrop",
                ),
            ),
        )

    def _add_measurement_uuids(self, data: pd.DataFrame) -> pd.DataFrame:
        data["a260 uuid"] = [random_uuid_str() for _ in range(len(data.index))]
        data["a280 uuid"] = [random_uuid_str() for _ in range(len(data.index))]
        return data

    def _get_spectrophotometry_document(
        self, data: pd.DataFrame
    ) -> list[SpectrophotometryDocumentItem]:
        return [
            self._get_spectrophotometry_document_item(data, i)
            for i in range(len(data.index))
        ]

    def _get_calculated_data_document(
        self, data: pd.DataFrame
    ) -> list[CalculatedDataDocumentItem]:
        calculated_data_documents = []
        for i in range(len(data.index)):
            if _get_str_or_none(data, i, "260/280"):
                calculated_data_documents.append(self._get_260_280(data, i))

            if _get_str_or_none(data, i, "260/230"):
                calculated_data_documents.append(self._get_260_230(data, i))

        return calculated_data_documents

    def _get_260_280(self, data: pd.DataFrame, row: int) -> CalculatedDataDocumentItem:
        data_source_doc_items = []
        if _get_str_or_none(data, row, "a260"):
            data_source_doc_items.append(
                DataSourceDocumentItem(
                    data_source_feature="absorbance",
                    data_source_identifier=_get_str(data, row, "a260 uuid"),
                )
            )

        if _get_str_or_none(data, row, "a280") or _get_str_or_none(
            data, row, "a280 10mm"
        ):
            data_source_doc_items.append(
                DataSourceDocumentItem(
                    data_source_feature="absorbance",
                    data_source_identifier=_get_str(data, row, "a280 uuid"),
                )
            )

        data_source_aggregate_document = None
        if data_source_doc_items:
            data_source_aggregate_document = DataSourceAggregateDocument(
                data_source_document=data_source_doc_items
            )

        return CalculatedDataDocumentItem(
            calculated_data_name="A260/280",
            calculated_result=TQuantityValue(
                value=_get_float(data, row, "260/280"), unit=UNITLESS
            ),
            calculated_data_identifier=random_uuid_str(),
            data_source_aggregate_document=data_source_aggregate_document,
        )

    def _get_260_230(self, data: pd.DataFrame, row: int) -> CalculatedDataDocumentItem:
        data_source_doc_items = []
        if _get_str_or_none(data, row, "a260"):
            data_source_doc_items.append(
                DataSourceDocumentItem(
                    data_source_feature="absorbance",
                    data_source_identifier=_get_str(data, row, "a260 uuid"),
                )
            )

        data_source_aggregate_document = None
        if data_source_doc_items:
            data_source_aggregate_document = DataSourceAggregateDocument(
                data_source_document=data_source_doc_items
            )
        return CalculatedDataDocumentItem(
            calculated_data_name="A260/230",
            calculated_result=TQuantityValue(
                value=_get_float(data, row, "260/230"), unit=UNITLESS
            ),
            calculated_data_identifier=random_uuid_str(),
            data_source_aggregate_document=data_source_aggregate_document,
        )

    def _get_spectrophotometry_document_item(
        self, data: pd.DataFrame, row: int
    ) -> SpectrophotometryDocumentItem:
        return SpectrophotometryDocumentItem(
            analyst=_get_str_or_none(data, row, "user id"),
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=self._get_date_time(
                    _get_str(data, row, "date") + " " + _get_str(data, row, "time")
                ),
                experiment_type=_get_str_or_none(data, row, "na type"),
                measurement_document=self._get_measurement_document(data=data, row=row),
            ),
        )

    def _get_measurement_document(
        self, data: pd.DataFrame, row: int
    ) -> list[UltravioletAbsorbancePointDetectionMeasurementDocumentItems]:
        measurement_docs = []
        na_type = _get_str_or_none(data, row, "na type")
        concentration_col = self._get_concentration_col(data)
        a280_col = "a280"
        if a280_col not in data.columns and "a280 10mm" in data.columns:
            a280_col = "a280 10mm"

        if _get_str_or_none(data, row, "a260"):
            # capture concentration on the A260 measurement document if the experiment type is
            # DNA or RNA, protein and other concentration is captured on A280 measurment
            # if there is no experiment type and no 280 column add the concentration here

            mass_concentration = None
            processed_data_aggregate_document = None

            if concentration_col and (
                (na_type is not None and "NA" in na_type)
                or (na_type is None and a280_col not in data.columns)
            ):
                mass_concentration = _get_concentration(
                    _get_float(data, row, concentration_col),
                    _get_str_or_none(data, row, "units"),
                )

            if mass_concentration is not None:
                processed_data_aggregate_document = ProcessedDataAggregateDocument(
                    processed_data_document=[
                        ProcessedDataDocumentItem(mass_concentration=mass_concentration)
                    ]
                )

            measurement_docs.append(
                UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                    measurement_identifier=_get_str(data, row, "a260 uuid"),
                    sample_document=SampleDocument(
                        sample_identifier=_get_str(data, row, "sample id")
                        if _get_str_or_none(data, row, "sample id")
                        else "NA",
                        well_plate_identifier=_get_str_or_none(data, row, "plate ID"),
                        location_identifier=_get_str_or_none(data, row, "well"),
                    ),
                    processed_data_aggregate_document=processed_data_aggregate_document,
                    device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                        device_control_document=[
                            UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                                device_type="absorbance detector",
                                detector_wavelength_setting=TQuantityValueNanometer(
                                    value=260
                                ),
                            )
                        ]
                    ),
                    absorbance=TQuantityValueMilliAbsorbanceUnit(
                        _get_float(data, row, "a260")
                    ),
                )
            )

        if _get_str_or_none(data, row, a280_col):
            # capture concentration on the A280 measurement document if the experiment type is
            # something other than DNA or RNA or if the experiment type is not specified
            mass_concentration = None
            if (na_type is not None and "NA" not in na_type and concentration_col) or (
                na_type is None and concentration_col
            ):
                mass_concentration = _get_concentration(
                    _get_float(data, row, str(concentration_col)),
                    _get_str_or_none(data, row, "units"),
                )
            processed_data_aggregate_document = None
            if mass_concentration:
                processed_data_aggregate_document = ProcessedDataAggregateDocument(
                    processed_data_document=[
                        ProcessedDataDocumentItem(
                            # capture concentration on the A280 measurement document if the experiment type is
                            # something other than DNA or RNA or n ot specified
                            mass_concentration=mass_concentration
                        )
                    ]
                )
            measurement_docs.append(
                UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                    measurement_identifier=_get_str(data, row, "a280 uuid"),
                    sample_document=SampleDocument(
                        sample_identifier=_get_str(data, row, "sample id")
                        if _get_str_or_none(data, row, "sample id")
                        else "NA",
                        well_plate_identifier=_get_str_or_none(data, row, "plate id"),
                        location_identifier=_get_str_or_none(data, row, "well"),
                    ),
                    processed_data_aggregate_document=processed_data_aggregate_document,
                    device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                        device_control_document=[
                            UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                                device_type="absorbance detector",
                                detector_wavelength_setting=TQuantityValueNanometer(
                                    value=280
                                ),
                            )
                        ]
                    ),
                    absorbance=TQuantityValueMilliAbsorbanceUnit(
                        _get_float(data, row, a280_col)
                    ),
                )
            )

        return measurement_docs

    def _get_concentration_col(self, data: pd.DataFrame) -> Optional[str]:
        for col in data.columns:
            if col.lower() in ["conc.", "conc", "concentration"]:
                return col
        return None



import numpy as np
import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.types import IOType


class AbsoluteQReader:
    def __init__(self, contents: IOType):
        absolute_q_data = read_csv(
            filepath_or_buffer=contents, parse_dates=["Date"]
        ).replace(np.nan, None)
        self.wells: pd.DataFrame = absolute_q_data.dropna(subset=["Name"])
        self.group_rows: pd.DataFrame = absolute_q_data[absolute_q_data["Name"].isna()]



import json
import os
from pathlib import Path
import re
import subprocess  # noqa: S404, RUF100
from typing import Optional

from autoflake import fix_file  # type: ignore[import-untyped]
from datamodel_code_generator import (
    DataModelType,
    generate,
    InputFileType,
    PythonVersion,
)

from allotropy.allotrope.schema_parser.backup_manager import (
    backup,
    is_file_changed,
    restore_backup,
)
from allotropy.allotrope.schema_parser.model_class_editor import modify_file
from allotropy.allotrope.schemas import get_schema

SCHEMA_DIR_PATH = "src/allotropy/allotrope/schemas"
MODEL_DIR_PATH = "src/allotropy/allotrope/models"


def lint_file(model_path: str) -> None:
    # The first run of ruff changes typing annotations and causes unused imports. We catch failure
    # due to unused imports.
    try:
        subprocess.check_call(
            f"ruff {model_path} --fix",
            shell=True,  # noqa: S602
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        pass
    # The call to autoflake.fix_file removes unused imports.
    fix_file(
        model_path,
        {
            "in_place": True,
            "remove_unused_variables": True,
            "write_to_stdout": False,
            "ignore_init_module_imports": False,
            "expand_star_imports": False,
            "remove_all_unused_imports": True,
            "remove_duplicate_keys": True,
            "remove_rhs_for_unused_variables": False,
            "ignore_pass_statements": False,
            "ignore_pass_after_docstring": False,
            "check": False,
            "check_diff": False,
        },
    )
    # The second call to ruff checks for additional rules.
    subprocess.check_call(
        f"ruff {model_path} --fix", shell=True, stdout=subprocess.DEVNULL  # noqa: S602
    )
    subprocess.check_call(
        f"black {model_path}", shell=True, stderr=subprocess.DEVNULL  # noqa: S602
    )


def _get_schema_and_model_paths(
    root_dir: Path, rel_schema_path: Path
) -> tuple[Path, Path]:
    schema_path = Path(root_dir, SCHEMA_DIR_PATH, rel_schema_path)
    model_file = re.sub(
        "/|-", "_", f"{rel_schema_path.parent}_{rel_schema_path.stem}.py"
    ).lower()
    model_path = Path(root_dir, MODEL_DIR_PATH, model_file)
    return schema_path, model_path


def _generate_schema(
    model_path: Path, schema_path: Path, rel_schema_path: Path
) -> None:
    # get_schema adds extra defs from shared definitions to the schema.
    schema = get_schema(str(rel_schema_path))
    with open(schema_path, "w") as f:
        json.dump(schema, f)

    # Generate models
    generate(
        input_=schema_path,
        output=model_path,
        output_model_type=DataModelType.DataclassesDataclass,
        input_file_type=InputFileType.JsonSchema,
        # Specify base_class as empty when using dataclass
        base_class="",
        target_python_version=PythonVersion.PY_39,
        use_union_operator=False,
    )
    # Import classes from shared files, remove unused classes, format.
    modify_file(str(model_path), str(schema_path))
    lint_file(str(model_path))


def generate_schemas(
    root_dir: Path,
    *,
    dry_run: Optional[bool] = False,
    schema_regex: Optional[str] = None,
) -> list[str]:
    """Generate schemas from JSON schema files.

    :root_dir: The root directory of the project.
    :dry_run: If true, does not save changes to any models, but still returns the list of models that would change.
    :schema_regex: If set, filters schemas to generate using regex.
    :return: A list of model files that were changed.
    """

    os.chdir(os.path.join(root_dir, SCHEMA_DIR_PATH))
    schema_paths = list(Path(".").rglob("*.json"))
    os.chdir(os.path.join(root_dir))
    models_changed = []
    for rel_schema_path in schema_paths:
        if rel_schema_path.parts[0] == "shared":
            continue
        if schema_regex and not re.match(schema_regex, str(rel_schema_path)):
            continue

        print(f"Generating models for schema: {rel_schema_path}...")  # noqa: T201
        schema_path, model_path = _get_schema_and_model_paths(root_dir, rel_schema_path)

        with backup(model_path, restore=dry_run), backup(schema_path, restore=True):
            _generate_schema(model_path, schema_path, rel_schema_path)

            if is_file_changed(model_path):
                models_changed.append(Path(model_path).stem)
            else:
                restore_backup(model_path)

    return models_changed



from __future__ import annotations

import pandas as pd

from allotropy.allotrope.models.pcr_benchling_2023_09_dpcr import (
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    DPCRAggregateDocument,
    DPCRDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
    TQuantityValueNumberPerMicroliter,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_reader import QiacuitydPCRReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    try_float_from_series,
    try_float_from_series_or_none,
    try_int_from_series,
    try_int_from_series_or_none,
    try_str_from_series,
    try_str_from_series_or_none,
)
from allotropy.parsers.vendor_parser import VendorParser

BRAND_NAME = "Qiacuity Digital PCR System"
PRODUCT_MANUFACTURER = "Qiagen"
SOFTWARE_NAME = "Qiacuity Software Suite"
DEVICE_TYPE = "dPCR"
DEVICE_IDENTIFIER = "Qiacuity dPCR"
EPOCH = "1970-01-01T00:00:00-00:00"

TARGET_COLUMN_NAME = "Target"
PARTITIONS_COLUMN_NAME = "Partitions (valid)"
SAMPLE_IDENTIFIER_COLUMN_NAME = "Sample/NTC/Control"
SAMPLE_TYPE_COLUMN_NAME = "Type"
WELL_COLUMN_NAME = "Well Name"
CONCENTRATION_COLUMN_NAME = "Concentration (copies/µL)"
POSITIVE_COUNT_COLUMN_NAME = "Partitions (positive)"
NEGATIVE_COUNT_COLUMN_NAME = "Partitions (negative)"
FIT_SETTING_COLUMN_NAME = "Threshold"
WELL_PLATE_IDENTIFIER_COLUMN_NAME = "Plate ID"

SAMPLE_ROLE_TYPE_MAPPING = {
    "Sample": "Sample Role",
    "Control": "Control Sample Role",
    "Non Template Control": "Blank Role",
}


class QiacuitydPCRParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = named_file_contents.contents
        reader = QiacuitydPCRReader(contents)

        filename = named_file_contents.original_file_name
        return self._get_model(reader.well_data, filename)

    def _get_model(self, qiacuity_data: pd.DataFrame, file_name: str) -> Model:
        return Model(
            dPCR_aggregate_document=DPCRAggregateDocument(
                device_system_document=self._get_device_system_document(),
                data_system_document=self._get_data_system_document(
                    file_name=file_name
                ),
                dPCR_document=[self._get_dpcr_document(qiacuity_data=qiacuity_data)],
            )
        )

    def _get_dpcr_document(self, qiacuity_data: pd.DataFrame) -> DPCRDocumentItem:
        measurement_documents = []
        for _, well_item in qiacuity_data.iterrows():
            sample_document = self._get_sample_document(well_item=well_item)
            device_control_documents = [self._get_device_control_document()]
            device_control_aggregate_document = DeviceControlAggregateDocument(
                device_control_document=device_control_documents
            )
            processed_data_aggregate_document = ProcessedDataAggregateDocument(
                processed_data_document=[
                    self._get_processed_data_document(well_item=well_item)
                ]
            )
            measurement_document = self._get_measurement_document(
                well_item=well_item,
                sample_document=sample_document,
                device_control_aggregate_document=device_control_aggregate_document,
                processed_data_aggregate_document=processed_data_aggregate_document,
            )
            measurement_documents.append(measurement_document)
        # TODO: Hardcoded plate well count to 0 since it's a required field
        #  ASM will be modified to optional in future version
        measurement_aggregate_document = MeasurementAggregateDocument(
            measurement_document=measurement_documents,
            plate_well_count=TQuantityValueNumber(value=0),
        )
        return DPCRDocumentItem(
            measurement_aggregate_document=measurement_aggregate_document
        )

    def _get_device_system_document(self) -> DeviceSystemDocument:
        device_system_document = DeviceSystemDocument(
            device_identifier=DEVICE_IDENTIFIER,
            brand_name=BRAND_NAME,
            product_manufacturer=PRODUCT_MANUFACTURER,
        )
        return device_system_document

    def _get_data_system_document(self, file_name: str) -> DataSystemDocument:
        data_system_document = DataSystemDocument(
            file_name=file_name,
            software_name=SOFTWARE_NAME,
            ASM_converter_name=ASM_CONVERTER_NAME,
            ASM_converter_version=ASM_CONVERTER_VERSION,
        )
        return data_system_document

    def _get_measurement_document(
        self,
        well_item: pd.Series[str],
        sample_document: SampleDocument,
        device_control_aggregate_document: DeviceControlAggregateDocument,
        processed_data_aggregate_document: ProcessedDataAggregateDocument,
    ) -> MeasurementDocumentItem:
        # Assign a random measurement id
        measurement_id = random_uuid_str()
        # There is no measurement time in the file, so assign to unix epoch
        measurement_time = EPOCH
        target_dna_description = try_str_from_series(well_item, TARGET_COLUMN_NAME)
        total_partition_count = TQuantityValueNumber(
            try_int_from_series(well_item, PARTITIONS_COLUMN_NAME)
        )
        return MeasurementDocumentItem(
            measurement_identifier=measurement_id,
            measurement_time=measurement_time,
            target_DNA_description=target_dna_description,
            total_partition_count=total_partition_count,
            sample_document=sample_document,
            device_control_aggregate_document=device_control_aggregate_document,
            processed_data_aggregate_document=processed_data_aggregate_document,
        )

    def _get_sample_document(self, well_item: pd.Series[str]) -> SampleDocument:
        sample_identifier = try_str_from_series(
            well_item, SAMPLE_IDENTIFIER_COLUMN_NAME
        )

        sample_document = SampleDocument(sample_identifier=sample_identifier)

        sample_role_type = try_str_from_series_or_none(
            well_item, SAMPLE_TYPE_COLUMN_NAME
        )
        # TODO: When the sample role type model is updated in this repo, we should update this
        # Map sample role types to valid sample role types from ASM
        if sample_role_type is not None:
            try:
                sample_role_type = SAMPLE_ROLE_TYPE_MAPPING[sample_role_type]
                sample_document.sample_role_type = sample_role_type
            except KeyError as e:
                error_message = (
                    f"Unexpected sample type found: {sample_role_type}. "
                    f"Must be one of {list(SAMPLE_ROLE_TYPE_MAPPING.keys())}"
                )
                raise AllotropeConversionError(error_message) from e

        well_location_identifier = try_str_from_series_or_none(
            well_item, WELL_COLUMN_NAME
        )

        if well_location_identifier is not None:
            sample_document.well_location_identifier = well_location_identifier

        well_plate_identifier = try_str_from_series_or_none(
            well_item, WELL_PLATE_IDENTIFIER_COLUMN_NAME
        )
        if well_plate_identifier is not None:
            sample_document.well_plate_identifier = well_plate_identifier
        return sample_document

    def _get_device_control_document(self) -> DeviceControlDocumentItem:
        return DeviceControlDocumentItem(
            device_type=DEVICE_TYPE, device_identifier=DEVICE_IDENTIFIER
        )

    def _get_processed_data_document(
        self, well_item: pd.Series[str]
    ) -> ProcessedDataDocumentItem:

        number_concentration = TQuantityValueNumberPerMicroliter(
            value=try_float_from_series(well_item, CONCENTRATION_COLUMN_NAME)
        )
        positive_partition_count = TQuantityValueNumber(
            value=try_int_from_series(well_item, POSITIVE_COUNT_COLUMN_NAME)
        )
        processed_data_document = ProcessedDataDocumentItem(
            number_concentration=number_concentration,
            positive_partition_count=positive_partition_count,
        )
        # If the fluorescence intensity threshold setting exists, create a data processing document for it and add to processed data document
        fluor_intensity_threshold = try_float_from_series_or_none(
            well_item, FIT_SETTING_COLUMN_NAME
        )
        if fluor_intensity_threshold is not None:
            data_processing_document = DataProcessingDocument(
                flourescence_intensity_threshold_setting=TQuantityValueUnitless(
                    fluor_intensity_threshold
                )
            )
            processed_data_document.data_processing_document = data_processing_document

        # Negative partition count is optional
        negative_partition_count = try_int_from_series_or_none(
            well_item, NEGATIVE_COUNT_COLUMN_NAME
        )

        if negative_partition_count is not None:
            processed_data_document.negative_partition_count = TQuantityValueNumber(
                negative_partition_count
            )
        return processed_data_document



from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.definitions import (
    TNullableQuantityValue,
    TQuantityValue,
)
from allotropy.allotrope.models.shared.definitions.units import (
    Cell,
    Centimeter,
    CubicMillimeter,
    DegreeCelsius,
    GramPerLiter,
    Hertz,
    MicrogramPerMicroliter,
    MicrogramPerMilliliter,
    Microliter,
    Micrometer,
    MilliAbsorbanceUnit,
    MilligramPerMilliliter,
    Millimeter,
    MillimeterOfMercury,
    MillimolePerLiter,
    MillionCellsPerMilliliter,
    MilliOsmolesPerKilogram,
    MilliSecond,
    NanogramPerMicroliter,
    NanogramPerMilliliter,
    Nanometer,
    Number,
    NumberPerMicroliter,
    OpticalDensity,
    Percent,
    PH,
    PicogramPerMilliliter,
    RelativeFluorescenceUnit,
    RelativeLightUnit,
    SecondTime,
    SquareCentimetersPerGram,
    TODO,
    Unitless,
    UnitPerLiter,
)


@dataclass
class TQuantityValueCentimeter(Centimeter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMillimeter(Millimeter, TQuantityValue):
    pass


@dataclass
class TQuantityValueNanometer(Nanometer, TQuantityValue):
    pass


@dataclass
class TQuantityValueMicrometer(Micrometer, TQuantityValue):
    pass


@dataclass
class TQuantityValueHertz(Hertz, TQuantityValue):
    pass


@dataclass
class TQuantityValueCubicMillimeter(CubicMillimeter, TQuantityValue):
    pass


@dataclass
class TQuantityValueUnitless(Unitless, TQuantityValue):
    pass


@dataclass
class TQuantityValueSecondTime(SecondTime, TQuantityValue):
    pass


@dataclass
class TQuantityValuePercent(Percent, TQuantityValue):
    pass


@dataclass
class TQuantityValueCell(Cell, TQuantityValue):
    pass


@dataclass
class TQuantityValueMicroliter(Microliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueNumberPerMicroliter(NumberPerMicroliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMilliSecond(MilliSecond, TQuantityValue):
    pass


@dataclass
class TQuantityValueMillionCellsPerMilliliter(
    MillionCellsPerMilliliter, TQuantityValue
):
    pass


@dataclass
class TQuantityValueTODO(TODO, TQuantityValue):
    pass


@dataclass
class TQuantityValueGramPerLiter(GramPerLiter, TQuantityValue):
    pass


@dataclass
class TQuantityValueDegreeCelsius(DegreeCelsius, TQuantityValue):
    pass


@dataclass
class TQuantityValueNumber(Number, TQuantityValue):
    pass


@dataclass
class TQuantityValueNanogramPerMicroliter(NanogramPerMicroliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMicrogramPerMicroliter(MicrogramPerMicroliter, TQuantityValue):
    pass


@dataclass
class TQuantityValuePicogramPerMilliliter(PicogramPerMilliliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueNanogramPerMilliliter(NanogramPerMilliliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMicrogramPerMilliliter(MicrogramPerMilliliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMilligramPerMilliliter(MilligramPerMilliliter, TQuantityValue):
    pass


@dataclass
class TQuantityValueUnitPerLiter(UnitPerLiter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMillimeterOfMercury(MillimeterOfMercury, TQuantityValue):
    pass


@dataclass
class TQuantityValueOpticalDensity(OpticalDensity, TQuantityValue):
    pass


@dataclass
class TQuantityValuePH(PH, TQuantityValue):
    pass


@dataclass
class TQuantityValueMilliOsmolesPerKilogram(MilliOsmolesPerKilogram, TQuantityValue):
    pass


@dataclass
class TQuantityValueMillimolePerLiter(MillimolePerLiter, TQuantityValue):
    pass


@dataclass
class TQuantityValueMilliAbsorbanceUnit(MilliAbsorbanceUnit, TQuantityValue):
    pass


@dataclass
class TQuantityValueRelativeFluorescenceUnit(RelativeFluorescenceUnit, TQuantityValue):
    pass


@dataclass
class TQuantityValueRelativeLightUnit(RelativeLightUnit, TQuantityValue):
    pass


@dataclass
class TQuantityValueSquareCentimetersPerGram(SquareCentimetersPerGram, TQuantityValue):
    pass


@dataclass
class TNullableQuantityValueCell(Cell, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValueGramPerLiter(GramPerLiter, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValueMicrometer(Micrometer, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValueMilliOsmolesPerKilogram(
    MilliOsmolesPerKilogram, TNullableQuantityValue
):
    pass


@dataclass
class TNullableQuantityValueMillimeterOfMercury(
    MillimeterOfMercury, TNullableQuantityValue
):
    pass


@dataclass
class TNullableQuantityValueMillimolePerLiter(
    MillimolePerLiter, TNullableQuantityValue
):
    pass


@dataclass
class TNullableQuantityValueMillionCellsPerMilliliter(
    MillionCellsPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass
class TNullableQuantityValueOpticalDensity(OpticalDensity, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValuePH(PH, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValuePercent(Percent, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValueTODO(TODO, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValueUnitPerLiter(UnitPerLiter, TNullableQuantityValue):
    pass


@dataclass
class TNullableQuantityValueUnitless(Unitless, TNullableQuantityValue):
    pass



from typing import Any, Optional

import pandas as pd
from pandas import Timestamp

from allotropy.allotrope.models.cell_counting_benchling_2023_11_cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument1,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    TDateTimeValue,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.chemometec_nucleoview.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    NUCLEOCOUNTER_SOFTWARE_NAME,
)
from allotropy.parsers.chemometec_nucleoview.nucleoview_reader import NucleoviewReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser

_PROPERTY_LOOKUP = {
    "Dead (cells/ml)": TQuantityValueMillionCellsPerMilliliter,
    "Estimated cell diameter (um)": TQuantityValueMicrometer,
    "Live (cells/ml)": TQuantityValueMillionCellsPerMilliliter,
    "Multiplication factor": TQuantityValueUnitless,
    "Total (cells/ml)": TQuantityValueMillionCellsPerMilliliter,
    "Viability (%)": TQuantityValuePercent,
}


def _get_value(data_frame: pd.DataFrame, row: int, column: str) -> Optional[Any]:
    if column not in data_frame.columns:
        return None
    return data_frame[column][row]


def get_property_from_sample(
    data_frame: pd.DataFrame, row: int, property_name: str
) -> Optional[Any]:
    value = _get_value(data_frame, row, property_name)
    if value is None:
        return None

    property_type = _PROPERTY_LOOKUP[property_name]

    try:
        value = float(value)
    except ValueError:
        return property_type(value=InvalidJsonFloat.NaN)

    # if the porperty type is measured in million cells per ml convert cells per ml
    if property_type == TQuantityValueMillionCellsPerMilliliter:
        return property_type(value=float(value) / 1e6)

    return property_type(value=float(value))


class ChemometecNucleoviewParser(VendorParser):
    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = named_file_contents.contents
        filename = named_file_contents.original_file_name
        return self._get_model(NucleoviewReader.read(contents), filename)

    def _get_model(self, data: pd.DataFrame, filename: str) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest",
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=_get_value(data, 0, "Instrument type")
                    or DEFAULT_MODEL_NUMBER,
                    equipment_serial_number=_get_value(data, 0, "Instrument s/n"),
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=NUCLEOCOUNTER_SOFTWARE_NAME,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=self._get_cell_counting_document(data),
            ),
        )

    def _get_cell_counting_document(
        self, data: pd.DataFrame
    ) -> list[CellCountingDocumentItem]:
        return [
            self._get_cell_counting_document_item(data, i)
            for i in range(len(data.index))
            if _get_value(data, i, "Total (cells/ml)")
        ]

    def _get_date_time_or_epoch(self, time_val: Optional[Timestamp]) -> TDateTimeValue:
        if time_val is None:
            # return epoch time 1970-01-01
            return self._get_date_time("1970-01-01")
        return self._get_date_time_from_timestamp(time_val)

    def _get_cell_counting_document_item(
        self, data_frame: pd.DataFrame, row: int
    ) -> CellCountingDocumentItem:
        return CellCountingDocumentItem(
            analyst=_get_value(data_frame, row, "Operator") or DEFAULT_ANALYST,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    CellCountingDetectorMeasurementDocumentItem(
                        measurement_identifier=random_uuid_str(),
                        measurement_time=self._get_date_time_or_epoch(
                            _get_value(data_frame, row, "datetime")
                        ),
                        sample_document=SampleDocument(
                            sample_identifier=str(
                                _get_value(data_frame, row, "Sample ID")
                            )
                        ),
                        device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItemModel(
                                    device_type="dark field imager (cell counter)",
                                    detection_type="dark field",
                                )
                            ]
                        ),
                        processed_data_aggregate_document=ProcessedDataAggregateDocument1(
                            processed_data_document=[
                                ProcessedDataDocumentItem(
                                    data_processing_document=DataProcessingDocument(
                                        cell_density_dilution_factor=get_property_from_sample(
                                            data_frame, row, "Multiplication factor"
                                        ),
                                    ),
                                    viability__cell_counter_=get_property_from_sample(  # type: ignore[arg-type]
                                        data_frame, row, "Viability (%)"
                                    ),
                                    viable_cell_density__cell_counter_=get_property_from_sample(  # type: ignore[arg-type]
                                        data_frame, row, "Live (cells/ml)"
                                    ),
                                    dead_cell_density__cell_counter_=get_property_from_sample(
                                        data_frame, row, "Dead (cells/ml)"
                                    ),
                                    total_cell_density__cell_counter_=get_property_from_sample(
                                        data_frame, row, "Total (cells/ml)"
                                    ),
                                    average_total_cell_diameter=get_property_from_sample(
                                        data_frame, row, "Estimated cell diameter (um)"
                                    ),
                                )
                            ]
                        ),
                    )
                ],
            ),
        )



# generated by datamodel-codegen:
#   filename:  ultraviolet-absorbance.json
#   timestamp: 2023-09-17T16:00:51+00:00

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from allotropy.allotrope.models.shared.components.plate_reader import (
    ProcessedDataAggregateDocument,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMicroliter,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValuePicogramPerMilliliter,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDatacube,
    TDateTimeValue,
    TStringValue,
)


class ContainerType(Enum):
    reactor = "reactor"
    controlled_lab_reactor = "controlled lab reactor"
    tube = "tube"
    well_plate = "well plate"
    differential_scanning_calorimetry_pan = "differential scanning calorimetry pan"
    qPCR_reaction_block = "qPCR reaction block"
    vial_rack = "vial rack"
    pan = "pan"
    reservoir = "reservoir"
    array_card_block = "array card block"
    capillary = "capillary"
    disintegration_apparatus_basket = "disintegration apparatus basket"
    jar = "jar"
    container = "container"
    tray = "tray"
    basket = "basket"
    cell_holder = "cell holder"


class ScanPositionSettingPlateReader(Enum):
    bottom_scan_position__plate_reader_ = "bottom scan position (plate reader)"
    scan_position_configuration__plate_reader_ = (
        "scan position configuration (plate reader)"
    )
    top_scan_position__plate_reader_ = "top scan position (plate reader)"


@dataclass
class DeviceSystemDocument:
    device_identifier: TStringValue
    model_number: TStringValue


@dataclass
class DeviceControlDocumentItem:
    device_type: Optional[TStringValue] = None
    shaking_configuration_description: Optional[TStringValue] = None
    detector_distance_setting__plate_reader_: Optional[TQuantityValueMillimeter] = None
    integration_time: Optional[TQuantityValueSecondTime] = None
    number_of_averages: Optional[TQuantityValueNumber] = None
    detector_gain_setting: Optional[TStringValue] = None
    scan_position_setting__plate_reader_: Optional[
        ScanPositionSettingPlateReader
    ] = None
    detector_carriage_speed_setting: Optional[TStringValue] = None
    detector_wavelength_setting: Optional[TQuantityValueNanometer] = None
    detector_bandwidth_setting: Optional[TQuantityValueNanometer] = None
    field_index: Optional[int] = None


@dataclass
class DeviceControlAggregateDocument:
    device_control_document: Optional[list[DeviceControlDocumentItem]] = None


@dataclass
class MeasurementDocumentItem:
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    data_cube: Optional[TDatacube] = None
    compartment_temperature: Optional[TQuantityValueDegreeCelsius] = None
    processed_data_aggregate_document: Optional[ProcessedDataAggregateDocument] = None
    mass_concentration: Optional[TQuantityValuePicogramPerMilliliter] = None


@dataclass
class MeasurementAggregateDocument:
    measurement_identifier: TStringValue
    plate_well_count: TQuantityValueNumber
    measurement_document: list[MeasurementDocumentItem]
    measurement_time: Optional[TDateTimeValue] = None
    analyst: Optional[TStringValue] = None
    analytical_method_identifier: Optional[TStringValue] = None
    experimental_data_identifier: Optional[TStringValue] = None
    experiment_type: Optional[TStringValue] = None
    container_type: Optional[ContainerType] = None
    well_volume: Optional[TQuantityValueMicroliter] = None
    device_system_document: Optional[DeviceSystemDocument] = None


@dataclass
class Model:
    measurement_aggregate_document: Optional[MeasurementAggregateDocument] = None
    manifest: str = "http://purl.allotrope.org/manifests/ultraviolet-absorbance/BENCHLING/2023/09/ultraviolet-absorbance.manifest"



from __future__ import annotations

from typing import Optional

import numpy as np
import pandas as pd

from allotropy.parsers.utils.values import (
    assert_not_empty_df,
    assert_not_none,
)
from allotropy.types import IOType


class DesignQuantstudioContents:
    def __init__(self, contents_io: IOType) -> None:
        raw_contents = pd.read_excel(contents_io, header=None, sheet_name=None)
        contents = {
            str(name): df.replace(np.nan, None) for name, df in raw_contents.items()
        }

        self.header = self._get_header(contents)
        self.data = self._get_data(
            self.header.size + 1,  # plus 1 for empty line between header and data
            contents,
        )

    def _get_header(self, contents: dict[str, pd.DataFrame]) -> pd.Series[str]:
        sheet = assert_not_none(
            contents.get("Results"),
            msg="Unable to find 'Results' sheet.",
        )

        data = {}
        for _, * (title, value, *_) in sheet.itertuples():
            if title is None:
                break
            data[str(title)] = None if value is None else str(value)

        header = pd.Series(data)
        header.index = header.index.str.strip()
        return header

    def _get_data(
        self, drop_n_columns: int, contents: dict[str, pd.DataFrame]
    ) -> dict[str, pd.DataFrame]:
        data_structure = {}
        for name, sheet in contents.items():
            data = sheet.iloc[drop_n_columns:].reset_index(drop=True)
            data.columns = pd.Index(data.iloc[0])
            data_structure[name] = data.drop(0)
        return data_structure

    def get_sheet_or_none(self, sheet_name: str) -> Optional[pd.DataFrame]:
        return self.data.get(sheet_name)

    def get_non_empty_sheet_or_none(self, sheet_name: str) -> Optional[pd.DataFrame]:
        sheet = self.get_sheet_or_none(sheet_name)
        return None if sheet is None or sheet.empty else sheet

    def get_sheet(self, sheet_name: str) -> pd.DataFrame:
        return assert_not_none(
            self.get_sheet_or_none(sheet_name),
            msg=f"Unable to find '{sheet_name}' sheet in file.",
        )

    def get_non_empty_sheet(self, sheet_name: str) -> pd.DataFrame:
        return assert_not_empty_df(
            self.get_sheet(sheet_name),
            msg=f"sheet '{sheet_name}' is empty.",
        )



from collections.abc import Collection
from typing import Any, Optional

_ERROR_MESSAGE = "message must not be empty"


class AllotropeConversionError(Exception):
    def __init__(self, message: str) -> None:
        if not message or not message.strip():
            raise ValueError(_ERROR_MESSAGE)
        super().__init__(message)


def msg_for_error_on_unrecognized_value(
    key: str, value: Any, valid_values: Optional[Collection[Any]] = None
) -> str:
    msg = f"Unrecognized {key}: '{value}'."
    if valid_values:
        msg += f" Only {sorted(valid_values)} are supported."
    return msg
